//
// basicchannel.cc
//
//	Source for Basic+ I/O
//
// Description:
//
//	Class to handle Basic channels.
//
//	This class is used in a program (usually as an array of
//	channels) to handle the BASIC I/O statements.
//
//	It tries to map as much of the work through C++ streams
//	as possible, so that converting to pure C++ will be
//	easier.
//
// Bugs
//
//	Errors need to be improved. Decent numbers need to be allocated.
//	Really nee to get a basic error number header file.
//	Just throowing the most common errors, without really isolating
//		the exact error that should be returned.
//
// History
//
//	12/06/1998 - Kevin Handy
//		Original version
//		And things ACTUALLY work first try!
//		And amazin (with file output) works with only a little effort
//
//	04/08/1999 - Kevin Handy
//		Some improvements to code to get initial work on eof
//		error in getline, and clean up errors so it doesn't
//		output -1 errors.
//
//	04/02/2000 - Kevin Handy
//		Handle 'KB:' in open.
//
//	04/03/2000 - Kevin Handy
//		Change input routines to use ',' as separator.
//
//	04/11/2000 - Kevin Handy
//		Added more error trapping
//		Throw EOF errors.
//

//
// Include files
//
#include <stdlib.h>
#include <string.h>
#include "basicfun.h"
#include "basicchannel.h"

//
// Global Variables
//
int Recount;

//
// Constructors
//
BChannel::BChannel()
{
	outstream = NULL;
	instream = NULL;
	ioflag = 0;
	cpos = 1;
}

BChannel::~BChannel()
{
}

//
// Open without a file name
//
int BChannel::Open()
{
	cerr << "OPEN without name not yet functional" << endl;
	BasicError error(-1);
	throw(error);
	return 0;
}

//
// Open when given a Basic String
//
int BChannel::Open(const BString& Name)
{
	return Open(Name.c_str());
}

//
// Open when given a char string
//
int BChannel::Open(const char* Name)
{
	//
	// Handle the very common case of 'KB:' specially
	//
	if (strcasecmp(Name, "KB:") == 0)
	{
		OpenZero();
		return 0;
	}

	switch(ioflag)
	{
	case 0:	// Both input and output
		{
			ifstream *inxstream = new ifstream(Name, ios::in | ios::out);
			if (inxstream->bad())
			{
				BasicError error(5);	// Can't find file
				throw(error);
			}
			instream = inxstream;
			outstream = new ostream(inxstream->rdbuf());
		}
		break;

	case BASIC_FORINPUT:
		instream = new ifstream(Name);
		if (instream->bad())
		{
			BasicError error(5);	// Can't find file
			throw(error);
		}
		break;

	case BASIC_FOROUTPUT:
		outstream = new ofstream(Name);
		if (outstream->bad())
		{
			BasicError error(10);	// Protection violation
			throw(error);
		}
		break;
	default:
		{
			BasicError error(10);	// Protection violation
			throw(error);
		}
	}
	cpos = 1;
	return 0;
}

int BChannel::ForInput()
{
	Close();
	ioflag = BASIC_FORINPUT;
	return 0;
}

int BChannel::ForOutput()
{
	Close();
	ioflag = BASIC_FOROUTPUT;
	return 0;
}

//
// Function to set up the special channel 0
//
int BChannel::OpenZero()
{
	outstream = &cout;
	instream = &cin;
	cpos = 1;
}


//
// RMS Access
//
int BChannel::Close()
{
	//
	// Close any open ostream's
	//
	if ((outstream != NULL) && (outstream != &cout))
	{
		delete outstream;
	}
	//
	// and any open istreams
	//
	if ((instream != NULL) && ((void*)instream != (void*)outstream) &&
		(instream != &cin))
	{
		delete instream;
	}
	outstream = NULL;
	instream = NULL;
	ioflag = 0;

	return 0;
}

int BChannel::Delete()
{
	cerr << "DELETE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return 0;
}

int BChannel::Find()
{
	cerr << "FIND not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return 0;
}

int BChannel::Get()
{
	cerr << "GET not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return 0;
}

int BChannel::Put()
{
	cerr << "PUT not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return 0;
}

int BChannel::Reset()
{
	cerr << "RESET not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return 0;
}

int BChannel::Unlock()
{
	cerr << "UNLOCK not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return 0;
}


//
// Inputline
//
int BChannel::getline(char* Output, int MaxChar)
{
	cpos = 1;
	instream->getline(Output, MaxChar);
	if (instream->eof())
	{
		throw(BasicError(11));	// End of file on device
	}
	return 0;
}


//
// Special
//
char* BChannel::BufferLoc()
{
	cerr << "BUFFERLOC not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
	return NULL;
}

void BChannel::SetRecordSize()
{
	cerr << "SETRECORDSIZE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetFileSize()
{
	cerr << "SETFILESIZE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetExtendSize()
{
	cerr << "SETEXTENDSIZE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetWindowSize()
{
	cerr << "SETWINDOWSIZE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetBlockSize()
{
	cerr << "SETBLOCKSIZE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetBucketSize()
{
	cerr << "SETBUCKETSIZE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetMode()
{
	cerr << "SETMODE not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetBuffer()
{
	cerr << "SETBUFFER not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetDefaultName()
{
	cerr << "SETDEFAULTNAME not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetKey()
{
	cerr << "SETKEY not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetTemporary()
{
	cerr << "SETTEMPORARY not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetContiguous()
{
	cerr << "SETCONTIGUOUS not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetRewind()
{
	cerr << "SETREWIND not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetSpan()
{
	cerr << "SETSPAN not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetOrginization()
{
	cerr << "SETORGANIZED not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetAccess()
{
	cerr << "SETACCESS not yet functional" << endl;
//	BasicError error(18);		// Illegal SYS usage
//	throw(error);
}

void BChannel::SetAllow(int flag)
{
	cerr << "SETALLOW not yet functional" << endl;
//	BasicError error(18);		// Illegal SYS usage
//	throw(error);
}

void BChannel::SetConnect()
{
	cerr << "SETCONNECT not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetMap()
{
	cerr << "SETMAP not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetRecordType()
{
	cerr << "SETRECORD not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::GoofyOpenStuff()
{
	cerr << "SETGOOFYOPENSTUFF not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

void BChannel::SetRecord(int)
{
	cerr << "SETRECORD not yet functional" << endl;
	BasicError error(18);		// Illegal SYS usage
	throw(error);
}

int BChannel::Ccpos()
{
	return cpos;
}


BChannel& endl(BChannel& outs)
{
	if (&outs == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	outs.GetOstream() << endl;
	outs.cpos = 1;
	return outs;
}

BChannel& operator<< (BChannel& os, const BString& bs)
{
	if (&os == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	*(os.outstream) << bs;
	os.cpos += bs.length();
	return os;
}

BChannel& operator<< (BChannel& os, const char* bs)
{
	if (&os == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	*(os.outstream) << bs;
	os.cpos += strlen(bs);
	return os;
}

BChannel& operator<< (BChannel& os, Dfloat bs)
{
	if (&os == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	if (bs >= 0.0)
	{
		*(os.outstream) << ' ';
		os.cpos++;
	}
	*(os.outstream) << bs << ' ';
	// FIXME: Need to correctly calculate new position
	os.cpos += 2;
	return os;
}

BChannel& operator<< (BChannel& os, Int32 bs)
{
	if (&os == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	if (bs >= 0)
	{
		*(os.outstream) << ' ';
		os.cpos++;
	}
	*(os.outstream) << bs << ' ';
	// FIXME: Need to correctly calculate new position
	os.cpos += 3;
	return os;
}

BChannel& operator<< (BChannel& os, Int bs)
{
	if (&os == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	if (bs >= 0)
	{
		*(os.outstream) << ' ';
		os.cpos++;
	}
	*(os.outstream) << bs << ' ';
	// FIXME: Need to correctly calculate new position
	os.cpos += 3;
	return os;
}

BChannel& operator<< (BChannel& os, char bs)
{
	if (&os == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	*(os.outstream) << bs;

	//
	// Calculate current cursor position
	//
	switch(bs)
	{
	case '\n':
		os.cpos = 1;
		break;
	case '\t':
		os.cpos += (8 - os.cpos % 8);
		break;
	default:
		os.cpos++;
		break;
	}
	return os;
}

BString BChannel::Tab(int x)
{
        if ((x - cpos) > 0)
        {
                return BString(' ', x - cpos);
        }
        else
        {
                return BString("");
        }
}
 
//
// >> BString
//
//	Grab a string (up to a comma, tab, or newline
BChannel& operator>> (BChannel& is, BString& bs)
{
	if (&is == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	if (is.instream->eof())
	{
		BasicError error(11);	// End of file on device
		throw(error);
	}
	string inputstr;
	is.cpos = 1;
	while(1)
	{
		char inchar;
		is >> inchar;
		switch(inchar)
		{
		case '\n':
		case ',':
		case '\t':
			bs = inputstr;
			return is;

		case '\r':
			break;

		default:
			inputstr += inchar;
			break;

		}
	}
}

BChannel& operator>> (BChannel& is, Int32& bs)
{
	BString instring;
	is >> instring;
	bs = atol(instring.c_str());		// Need to check for errors here
	return is;
}

BChannel& operator>> (BChannel& is, Int& bs)
{
	BString instring;
	is >> instring;
	bs = atoi(instring.c_str());		// Need to check for errors here
	return is;
}

BChannel& operator>> (BChannel& is, Dfloat& bs)
{
	BString instring;
	is >> instring;
	bs = atof(instring.c_str());		// Need to check for errors here
	return is;
}

//
// >> char
//
//	Get a single character from the input stream
//
//	This routine is unlikely to be used in useer code, but is used
//	as a basis for all of the other input routines for BChannel
//
BChannel& operator>> (BChannel& is, char& bs)
{
	if (&is == NULL)
	{
		BasicError error(9);	// I/O channel not open
		throw(error);
	}
	if (is.instream->eof())
	{
		BasicError error(11);	// End of file on device
		throw(error);
	}
	(is.instream)->get(bs);		// Must use get() to be able to see \n
	is.cpos = 1;
	return is;
}

