//
// b1filter,cc
//
//	Filter old RSTS/E Basic+ program into VAX Basic format
//
// Description:
//
//	This filter will read in old format Basic Plus programs
//	and convert them into a more modern format that is more
//	acceptable to VAX Basic (and my Basic to C++ translator)
//
// Bugs:
//
//	Data statements and remark statements may confuse this
//	simple program, so watch them.
//
//	Its simple lexing routines can be easily confused, so
//	you may have to do some hand editing afterwards (yuck)
//
// History
//
//	03/13/2000 - Kevin Handy
//		Original version
//
//	03/17/2000 - Kevin Handy
//		Make compiler handle missing semicolons in print
//		statements, so I killed my poor hack that tried to fix
//		the problem here (and didn't work anyway).
//
//	03/31/2000 - Kevin Handy
//		Fix IsQuoted function to return the quote character
//		instead of bool.
//		Close quotes at the end of a line if they remain open.
//
//	07/25/2000 - Kevin Handy
//		More keywords added
//

#include "config.h"
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string>
#include <vector>

extern "C"
{
#include <getopt.h>
#include <ctype.h>
}

//
// Set up for the options
//
static const char* OptionList = "o:O:hHe";
#ifdef USE_LONGOPT
static const struct option OptionListLong[] =
{
	{"output", 1, 0, 'o'},
	{"help", 0, 0, 'h'},
	{"extended", 0, 0, 'e'},
	{0, 0, 0, 0}
};
#endif

//
// Local functions
//
static void ProcessFile(ostream* Outfile, istream* InFile, const char* name);
static void LoadFile(istream* InFile, vector<string> &source);
static int LoadLine(istream *Infile, string& oneline);
static void MungFile(vector<string> &source);
static void MungWhitespace(vector<string> &source);
static void MungWhitespaceLine(string &sloop);
static void MungContinuation(vector<string> &source);
static void Writefile(ostream *OutFile, vector<string> &source);
static int IsQuoted(const string& source, int pos);
static int isbpunct(char source);

//
// Global variables
//
int Extended = 0;

////////////////////////////////////////////////////////////////////
// Here is the main program
//
int main(int argc, char* argv[])
{
        int OptionSelected;		// getopt parameter
        int OptionIndex;		// getopt parameter
	int DidOneFlag = 0;		// Have we done any files?
	ostream *OutFile = &cout;	// Output file
	istream *InFile;		// Input file

        //
        // Handle input arguments
        //
#ifdef USE_LONGOPT
        while ((OptionSelected = getopt_long(argc, argv, OptionList,
                OptionListLong, &OptionIndex)) != -1)
#else
        while ((OptionSelected = getopt(argc, argv, OptionList)) != -1)
#endif
        {
                //
                // Process one arguement
                //
                switch (OptionSelected)
                {
		case 'o':
		case 'O':
			//
			// Output File
			//
			if (OutFile != &cout)
			{
				delete OutFile;
				OutFile = 0;
			}

			OutFile = new ofstream(optarg);
			if (OutFile->bad())
			{
				cerr << "Unable to open " << optarg <<
					"for output" << endl;
				exit(1);
			}

			break;

		case 'h':
		case 'H':
			DidOneFlag = 1;
			cerr << "Usage: basic [-h] "
				"[-o <file>] <source>" << endl;

			break;

		case 'e':
			Extended = 1;
			break;

		}
	}

	//
	// Anything left over is files names to process
	//
	while (optind < argc)
	{
		//
		// At least one file name on command line
		//
		DidOneFlag = 1;

		InFile = new ifstream(argv[optind]);
		if (InFile->bad())
		{
			cerr << argv[0] << ": Bad Input File " <<
				argv[optind] << endl;
		}
		else
		{
			ProcessFile(OutFile, InFile, argv[optind]);
			delete InFile;
		}

		optind++;
	}

	//
	// If we haven't seen any file names, then get input
	// from the users terminal
	//
	if (DidOneFlag == 0)
	{
		ProcessFile(OutFile, &cin, "Standard Input");
	}
}

////////////////////////////////////////////////////////////////////
// Handle everything for one file
//
static void ProcessFile(ostream* OutFile, istream* InFile, const char* name)
{
//	cout << "Processing  " << name << endl;

	vector<string> source;

	LoadFile(InFile, source);
	MungFile(source);
	Writefile(OutFile, source);
}

////////////////////////////////////////////////////////////////////
// Read in the file
//
static void LoadFile(istream* InFile, vector<string> &source)
{
	string oneline;

	//
	// Just keep looping until we see the end of file marker
	//
	while (LoadLine(InFile, oneline) == 0)
	{
		source.push_back(oneline);
	}
}

////////////////////////////////////////////////////////////////////
// Read in a single line, stripping off any garbage
//
//	Assumes that EOF will occur on a blank line
//
static int LoadLine(istream *InFile, string& oneline)
{
	oneline = "";
	int inchar;
	int squote = 0;
	int dquote = 0;

	while ((inchar = InFile->get()) != EOF)
	{
		switch(inchar)
		{
		case '&':
			//
			// Shorthand for print maybe
			//
			if ((squote == 0) && (dquote == 0) && (Extended == 0))
			{
				oneline += "PRINT";
			}
			else
			{
				if ((Extended == 0) || (squote != 0) ||
					(dquote != 0))
				{
					oneline += char(inchar);
				}
			}
			break;

		case ':':
			//
			// Old statement separator
			//
			if ((squote == 0) && (dquote == 0) && (Extended == 0))
			{
				oneline += "\\";
			}
			else
			{
				oneline += char(inchar);
			}
			break;

		case '"':
			//
			// Maybe start of a double quote
			//
			if (dquote)
			{
				dquote = 0;
			}
			else
			{
				if (squote == 0)
				{
					dquote = 1;
				}
			}
			oneline += char(inchar);
			break;

		case '\'':
			//
			// Maybe start of a single quote
			//
			if (squote)
			{
				squote = 0;
			}
			else
			{
				if (dquote == 0)
				{
					squote = 1;
				}
			}
			oneline += char(inchar);
			break;

		case '\r':
		case '\0':
			//
			// Lose stupid extra characters
			//
			break;

		case '\n':
			//
			// Aha! We have the actual end of a line
			//
			return 0;
			break;

		default:
			//
			// Just stuff it on the end of the string
			//
			if (squote || dquote)
			{
				oneline += char(inchar);
			}
			else
			{
				oneline += toupper(inchar);
			}
			break;
		}
	}

	return 1;
}

////////////////////////////////////////////////////////////////////
// Process the file into the correct format
//
static void MungFile(vector<string> &source)
{
	MungWhitespace(source);
	MungContinuation(source);
}

////////////////////////////////////////////////////////////////////
// Add spaces where necessary
//
//	This routine attempts to insert whitespace between common
//	basic commands.
//
static const char* reserved[] =
{
	"ASCII",
	"CLOSE",
	"CONTINUE",
	"DATA",
	"DEF",
	"DIM",
	"ELSE",
	"ERROR",
	"FILE",
	"FOR",
	"GOTO",
	"GO TO",
	"IF",
	"INPUT",
	"LET",
	"NEXT",
	"OPEN",
	"PRINT",
	"READ",
	"RECORDSIZE",
	"RECORD",
	"REM",
	"RESTORE",
	"THEN",
	"SIZE",
	"STOP",
	"STEP",
	"GO SUB",
	"GOSUB",
	"TO",
	"ON",
	"OR",
	"AS",
	0
};

static void MungWhitespace(vector<string> &source)
{
	for (vector<string>::iterator sloop = source.begin();
		sloop != source.end(); sloop++)
	{
		MungWhitespaceLine(*sloop);
	}
}

static void MungWhitespaceLine(string &source)
{
	for (int pos = 0; pos < source.length(); )
	{
		//
		// Are we looking at a reserved word at this position?
		//
		for (int rloop = 0; reserved[rloop] != 0; rloop++)
		{
			int rlen = strlen(reserved[rloop]);

			//
			// Do a really quick check on the first letter
			// followed by a full length match
			//
			if ((source[pos] == reserved[rloop][0]) &&
				(source.substr(pos, rlen) == reserved[rloop]))
			{
				//
				// Put spaces before string
				//
				if ((pos > 0) &&
					!isspace(source[pos - 1]) &&
					!isbpunct(source[pos - 1]))
				{
					source.insert(pos, " ");
					pos++;
				}

				//
				// Put spaces after string
				//
				if (!isspace(source[pos + rlen]) &&
					!isbpunct(source[pos + rlen]))
				{
					source.insert(pos + rlen, " ");
					pos++;
				}
				pos += rlen;
				goto endloop;
			}
		}

		//
		// Skip over single quote
		//
		if (source[pos] == '\'')
		{
			//
			// Search for closing quote
			//
			int fpos = source.find('\'', pos + 1);
			if (fpos != source.npos)
			{
				pos = fpos + 1;
				goto endloop;
			}

			//
			// Didn't find one? Add it in!
			//
			source += '\'';
			return;
		}

		//
		// Skip over double quote
		//
		if (source[pos] == '"')
		{
			//
			// Search for closing quote
			//
			int fpos = source.find('"', pos + 1);
			if (fpos != source.npos)
			{
				pos = fpos + 1;
				goto endloop;
			}

#ifdef DEBUG
cerr << "Appending quote to '" << source << "'" << pos <<
	"," << source.length() << endl;
#endif
			//
			// Didn't find one? Add it in!
			//
			source += '"';
			return;
		}

		//
		// Point to next character
		//
		pos++;
endloop:;
	}
}

static void MungWhitespaceLineOld(string &sloop)
{
	for (int rloop = 0; reserved[rloop] != 0; rloop++)
	{
		int fpos;
		while ((fpos = sloop.find(reserved[rloop], fpos)) !=
			 sloop.npos)
		{
			if (IsQuoted(sloop, fpos))
			{
				//
				// Ignore in quoted string
				//
				fpos++;
			}
			else
			{
				//
				// Put spaces before string
				//
				if ((fpos > 0) &&
					!isspace(sloop[fpos - 1]) &&
					!isbpunct(sloop[fpos - 1]))
				{
					sloop.insert(fpos, " ");
					fpos++;
				}

				//
				// Put spaces after string
				//
				int rlen = strlen(reserved[rloop]);
				if ((fpos > 0) &&
					!isspace(sloop[fpos + rlen]) &&
					!isbpunct(sloop[fpos + rlen]))
				{
					sloop.insert(fpos  + rlen, " ");
				}
				fpos += rlen;
			}
		}
	}
}

////////////////////////////////////////////////////////////////////
// Add continuation markers where necessary
//
static void MungContinuation(vector<string> &source)
{
	vector<string>::iterator previous = 0;
	for (vector<string>::iterator loop = source.begin();
		loop != source.end(); loop++)
	{
#if 0
		//
		// If there is an unterminated quote, then terminate it
		//
		int qw = IsQuoted(*loop, 30000);
		if (qw)
		{
			(*loop) += char(qw);
		}
#endif
		//
		// Check for a line number on the current line
		//
		if ((previous != 0) && (isdigit((*loop)[0]) == 0))
		{
			(*previous) += " &";
		}

		previous = loop;
	}
}

////////////////////////////////////////////////////////////////////
// Output the final result
//
static void Writefile(ostream *OutFile, vector<string> &source)
{
	//
	// Write out all of the strings that we have sucked in
	//
	for (vector<string>::iterator loop = source.begin();
		loop != source.end(); loop++)
	{
		*OutFile << *loop << endl;
	}
}

////////////////////////////////////////////////////////////////////
// Is this character inside of quotes?
//
static int IsQuoted(const string& source, int pos)
{
	int flag = 0;

	//
	// For the moment, assume it is not
	//
	for (int loop = 0; loop < source.length(); loop++)
	{
		int ch = source[loop];

		switch(ch)
		{
		//
		// Handle quote characters
		//
		case '"':
		case '\'':
			if (flag == 0)
			{
				flag = ch;
			}
			else
			{
				if (flag == ch)
				{
					flag = 0;
				}
			}
			break;
		}

		//
		// Have we reached the right position?
		//
		if (--pos <= 0)
		{
			return flag;
		}
	}

	//
	// Hit the end of the string before reaching the limiting position
	//
	return flag;
}

//
// isbpunct
//
//	Punctuation in basic that does not need whitespace arount it.
//
// Description:
//
//	This function exists (instead of using punct) because I need
//	to have the program add spaces around decimal points.
//
static int isbpunct(char source)
{
	switch(source)
	{
	case '"':
	case '\'':
	case ';':
	case '+':
	case '-':
	case '/':
	case '*':
		return 1;
	default:
		return 0;
	}
}
