/*******************************************************************************

Program:

	nodes3.cc

Title:

	Routines to work with nodes (code generation).

Description:

	This file contains various routines for manipulating the tree
	nodes used when parsing a basic program.

	This file contains the code generation routines.

Compile:

	cc [noopt]

History:

	07/08/1999 - Kevin Handy
		Split into three parts

	08/11/1999 - Kevin Handy
		Make Indent return a string, and merge it into the
		various 'os <<' streams instead as a seperate call.

	02/28/2000 - Kevin Handy
		Handle list of channels in closee statements.

	03/09/2000 - Kevin Handy
		Modify to create "basic_main" instead of "main" fucntion.

	03/09/2000 - Kevin Handy
		Use "BasicChannel[0]" instead of "cin" and "cout" so that
		all I/O changes can be done in one place.

	03/09/2000 - Kevin Handy
		Output 'Int' instead of 'int' for compatibility with
		everything else output.

	03/10/2000 - Kevin Handy
		Split function OutputIPData into OutputPrintData and
		OutputInputData to make it work properly.

	04/07/2000 - Kevin Handy
		Output code for "CHANGE"

	04/19/2000 - Kevin Handy
		Don't write out junk for "VIRTUAL" open.

	06/16/2000 - Kevin Handy
		Fix missing 'break' in 'OutputForcedType' function
		throw an error when a ON-GOTO or ON-GOSUB goes beyond bounds.

	06/28/2000 - Kevin Handy
			Fix OTHERWISE handling for ON statements.
			ON GOTO cases don't need to be followed by a break.

	07/01/2000 - Kevin Handy
		Split OutputCodeOne into OutputCodeOne and Expression.
		Many changes to make other functions return 'string' instead
		of directly outputing to an ostream.

	07/03/2000 - Keevin Handy
		Make more functions return strings instead of passing an
		ostream tyhrough.

	07/10/2000 - Kevin Handy
		Lose unused several unused functions, clean up some code and comments.
		Replace cases that should no longer occur with asserts.

	08/03/2000 - Kevin Handy
		Output dummy code for MAT operations.

	08/10/2000 - Kevin Handy
		Fill in a lot of missing code generation.

	09/27/2000 - Kevin Handy
		Lose spurious 'os << os' in RECORD.
		Use Block[1] instead of Tree[0] for record structure.
		Use Expression instead of OutputCodeOne in constant definitions.
		Change a many cases os "<< Tree[n]" to "Tree[n]->Expression()".

	10/10/2000 - Kevin Handy
		CASE for a VARIANT is Block[1] not Tree[3].

	10/13/2000 - Kevin Handy
		Create function OutputNodeVarType in order to handle structure names.

	10/18/2000 - Kevin Handy
		Fix code generated by 'UNLESS x%'

	10/18/2000 - Kevin Handy
		Make 'EXTERNAL SUB' so it doesn't crash compiler.

*******************************************************************************/

//
// System Include Files
//
#include <iostream.h>
#include <stdio.h>
#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

//
// Project Include Files
//
#include "basic.h"
#include "parse.h"
#include "nodes.h"
#include "varlist.h"

//
// Module Variables
//
extern int GosubFlag;
extern int OnErrorFlag;

static int IOChPrinted;		// Has channel number been output
static Node* IOChannel;		// IO Channel to use (NULL if default)
static Node* IOUsing;		// Print using format

/*******************************************************************************

Member Function:

	Node::Output()

Title:

	Main function to output translated code.

Description:

	This function outputs the translated code.

Bugs:

	Ugly Code.

*******************************************************************************/

void Node::Output(ostream& os)
{
	//
	// Handle Include Files
	//
	if (NeedIostreamH != 0)
	{
		os << "#include <iostream.h>" << endl;
		NeedIostreamH = 0;
	}

	os << "#include <stdlib.h>" << endl;
	NeedStdlibH = 0;

	os << "#include <unistd.h>" << endl;
	NeedUnistd = 0;

	if ((NeedStringH != 0) | (NeedBasicFun != 0))
	{
		os << "#include <string.h>" << endl;
		NeedStringH = 0;
	}

	if (NeedMathH != 0)
	{
		os << "#include <math.h>" << endl;
		NeedMathH = 0;
	}

	if (NeedErrorH != 0)
	{
		os << "#include <errno.h>" << endl;
		NeedErrorH = 0;
	}

	if (NeedTimeH != 0)
	{
		os << "#include <time.h>" << endl;
		NeedTimeH = 0;
	}

	os << "#include \"basicfun.h\"" << endl;
	os << "#include \"bstring.h\"" << endl;
	os << "#include \"basicchannel.h\"" << endl;

	NeedChannel = 0;
	NeedBasicFun = 0;

	if (NeedDataList != 0)
	{
		os << "#include \"datalist.h\"" << endl;
		NeedDataList = 0;
	}

	if (NeedPuse != 0)
	{
		os << "#include <pusing.h>" << endl;
	}

	if (NeedVirtual != 0)
	{
		os << "#include <virtual.h>" << endl;
		NeedVirtual = 0;
	}

	os << endl;

	if (NeedRFA != 0)
	{
		os << "typedef char RmsRfa[6];" << endl;
		NeedRFA = 0;
	}

	if (NeedPuse != 0)
	{
		os << "static PUsing PUse;\t// Needed for print using" << endl;
		NeedPuse = 0;
	}

	os << endl;

	//
	// Output the variables
	//
	Variables->OutputDef(os, Level);

	//
	// Dump out prototypes
	//
	OutputPrototypes(os);

	//
	// Output code
	//
	Level = 0;
	OutputCode(os);
}


/*******************************************************************************

Local Function:

	Node::OutputCode()

Title:

	Output one block of code.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

void Node::OutputCode(ostream& os)
{
	assert(this != NNULL);

	//
	// Local Variables
	//
	Node *Program = this;

	//
	// Starting at the top node, we scan down handling each
	// level 0 node type as necessary.
	//
	// Could do this by calling OutputCode recursively for the Down
	// level instead of splitting into OutputCode and OutputCodeOne,
	// but that would consume huge amounts of stack space.
	//
	while (Program != NNULL)
	{
		//
		// Handle Statement. Don't output code from included files.
		//
		if ((Program->FromInclude == 0) || (CompileFlag != 0))
		{
			Program->OutputCodeOne(os);
		}

		//
		// Get next node
		//
		Program = Program->Block[0];
	}
}


/*******************************************************************************

Local Function:

	Node::OutputBlock()

Title:

	Output one block of code.

Description:

	This function will output a block of code (if it exists)
	surrounded by parentheses, otherwise it will output a semicolon.

Bugs:


*******************************************************************************/

void Node::OutputBlock(ostream& os)
{
	//
	// If this is a NNULL bit of code, just output a semicolon
	//
	if (this == NNULL)
	{
		os << Indent() << ";" << endl;
	}
	else
	{
		//
		// Start the block
		//
		os << Indent() << "{" << endl;

		Level++;

		//
		// Local Variables
		//
		Node *Program = this;

		//
		// Starting at the top node, we scan down handling each
		// level 0 node type as necessary.
		//
		// Could do this by calling OutputCode recursively for the Down
		// level instead of splitting into OutputCode and OutputCodeOne,
		// but that would probibly consume too much stack space.
		//
		while (Program != NNULL)
		{
			//
			// Handle Statement
			//
			if ((Program->FromInclude == 0) || (CompileFlag != 0))
			{
				Program->OutputCodeOne(os);
			}

			//
			// Get next node
			//
			Program = Program->Block[0];
		}

		//
		// End the block
		//
		Level--;
		os << Indent() << "}" << endl;
	}
}


/*******************************************************************************

Local Function:

	Node::NoParen()

Title:

	Return a node losing any parentheses around the whole group.

Description:

	Removes the parentheses from around the nodes when they are
	not needed.

Bugs:

	This should probibly be taken care of during the parsing phase,
	which would simplify code generation a lot, but...

*******************************************************************************/

string Node::NoParen()
{
	if (Type == '(')
	{
		return Tree[0]->NoParen();
	}
	else
	{
		return Expression();
	}
}

/*******************************************************************************

Local Function:

	Node::OutputPrototypes()

Title:

	Output function Prototypes.

Description:

	This function scans through the basic blocks and dumps out
	function prototypes for all functions defined therin.

Bugs:


*******************************************************************************/

void Node::OutputPrototypes(ostream& os)
{
	Node* ThisNode = this;

	//
	// Output title for prototype section
	//
	os << endl << "//" << endl << "// Function Prototypes" << endl <<
		"//" << endl;

	//
	// Go through the level 0 functions
	//
	while(ThisNode != NNULL)
	{

		switch(ThisNode->Type)
		{
		case BAS_S_FUNCTION:
		case BAS_S_SUB:
		case BAS_S_DEF:
		case BAS_S_DEFSTAR:
		case BAS_S_HANDLER:

			//
			// Empty variable list
			//
			Variables->NewLevel();

			//
			// Handle any DEF*'s that are in there
			//
			if (ThisNode->Block[2] != NNULL)
			{
				ThisNode->Block[2]->OutputPrototypes(os);
			}

			//
			// Scan the call parameters
			//
			ThisNode->VariableScanOne(1);
			Variables->Fixup();

			//
			// Output function name
			//
			os << ThisNode->OutputNewDefinition() << ";" << endl;

			Variables->KillLevel();

			break;

		case BAS_S_MAINFUNCTION:
			//
			// Handle any DEF*'s that are in there
			//
			if (ThisNode->Block[2] != NNULL)
			{
				ThisNode->Block[2]->OutputPrototypes(os);
			}
			break;

		}

		ThisNode = ThisNode->Block[0];
	}

	os << endl;
}

/*******************************************************************************

Local Function:

	Node::OutputCodeOne()

Title:

	Output one block of code.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

void Node::OutputCodeOne(ostream& os)
{
	VariableStruct *ThisVar;
	Node* LookDown;
	int KeepGosubFlag;
	int KeepOnErrorFlag;

	assert(this != NNULL);
	assert(Type != BAS_S_CHANGE);

	//
	// Handle this node type
	//
	switch (Type)
	{
	case BAS_N_ASSIGN:
		os << Indent();
		if (Tree[0] != NNULL)
		{
			//
			// Function return values are stashed in the
			// 'Result' variable.
			//
			if (Tree[0]->Type == BAS_V_FUNCTION)
			{
				os << "Result = " << Tree[1]->Expression() << ";" << endl;
			}
			else
			{
				//
				// A simple assignment statement
				//
				os << Tree[0]->Expression() << " = " << Tree[1]->Expression() <<
					";" << endl;
			}
		}
		else
		{
			os << "Result = " << Tree[1]->Expression() << ";" << endl;
		}
		break;

	case BAS_S_CALL:
		os << Indent() << Tree[0]->Expression() << ";" << endl;
		break;

	case BAS_N_CAUSEERROR:
		os << Indent() <<
			"throw BasicError(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_CASE:
		if (Tree[0] != NNULL)
		{
			Tree[0]->OutputCaseLabel(os);
			os << endl;
		}
		else
		{
			os << Indent() << "case:\t// Empty case label" <<
				endl << endl;
		}

		//
		// Output any info under case statement
		//
		if (Block[1] != NNULL)
		{
			Block[1]->OutputCode(os);
		}
		os << Indent() << "break;" << endl << endl;
		break;

	case BAS_N_CASEELSE:
		Level--;
		os << Indent() << "default:" << endl;
		Level++;

		//
		// Output any info under case statement
		//
		if (Block[1] != NNULL)
		{
			Block[1]->OutputCode(os);
		}
		os << Indent() << "break;" << endl << endl;
		break;

	case BAS_P_ABORT:
		os << "#abort " << Tree[0]->Expression() << endl;
		break;

	case BAS_P_CDD:
		os << "#dictionary " << Tree[0]->Expression() << endl;
		break;

	case BAS_P_PRINT:
		os << "#message " << Tree[0]->Expression() << endl;
		break;

	case BAS_S_CHAIN:
		os << Indent() << "BasicChain(" <<
			Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_S_CHANGE1:
		//
		// Look for variable in table
		//
		ThisVar = Variables->Lookup(Tree[1]->TextValue,
			Tree[1]->Tree[0]);
		os << Indent() << "BChange1(" <<
			Tree[0]->Expression() << ", " << ThisVar->GetName() <<
			");" << endl;
		break;

	case BAS_S_CHANGE2:
		ThisVar = Variables->Lookup(Tree[0]->TextValue,
			Tree[0]->Tree[0]);
		os << Indent() << "BChange2(" <<
			ThisVar->GetName() << ", " << Tree[1]->Expression() <<
			");" << endl;
		break;

	case BAS_S_CLOSE:
		Tree[0]->OutputClose(os);
		break;

	case BAS_S_COMMON:
		OutputMap(os);
		break;

	case BAS_S_DATA:
		os << Indent() << "//" << endl <<
			Indent() << "// Data Statement" << endl <<
			Indent() << "//" << endl <<
			Indent() << "static const char* DataValue[] = {" <<
			endl;
		Level++;
		OutputData(os);
		os << ", NULL};" << endl;
		Level--;
		os << Indent() << "DataListClass DataList(DataValue);" << endl;
		break;

	case BAS_S_DECLARE:
		assert(Tree[0] != NNULL);
		Tree[0]->OutputDefinitionList(os, Tree[1], Type);
		break;

	case BAS_V_DECLARECONSTANT:
		assert(Tree[0] != NNULL);
		Tree[0]->OutputDefinitionList(os, NNULL, 0, 1);
		break;

	case BAS_V_DECLAREFUN:
		assert(Tree[0] != NNULL);
		Tree[0]->OutputDefinitionList(os, NNULL, 0, 2);
		break;

	case BAS_S_DELETE:
		os << Indent() << "BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Delete();" << endl;
		break;

	case BAS_S_SCRATCH:
		os << Indent() << "BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Scratch();" << endl;
		break;

	case BAS_S_DIM:
		assert(Tree[0] != NNULL);
		if (Tree[1] == NNULL)
		{
			Tree[0]->OutputDefinitionList(os, NNULL, 0);
		}
		else
		{
			Tree[0]->OutputVirtualList(os, Tree[1], NNULL, 0);
		}
		break;

	case BAS_S_ELSE:
	case BAS_S_USE:
	case BAS_S_END:
	case BAS_S_FNEND:
	case BAS_S_OPTION:
	case BAS_N_NULL:
		//
		// These nodes are not useful for code generation at
		// this point
		//
		break;

	case BAS_S_EXTERNAL:
		Tree[0]->OutputDefinitionList(os, NNULL, Type, 4);
		break;

	case BAS_N_EXTERNALSUB:
		os << Indent() << OutputNewDefinition() << ";" << endl;
		break;

	case BAS_N_EXTERNALCONSTANT:
		Tree[0]->OutputDefinitionList(os, Tree[1], Type, 3);
		break;

	case BAS_N_EXTERNALFUNCTION:
		os << Indent() << OutputNewDefinition() << ";" << endl;
		break;

	case BAS_S_EXIT:
		os << Indent() <<  "return;" << endl;
		break;

	case BAS_N_EXITHANDLER:
		// Re-throw the exception
		os << Indent() << "throw Be;" << endl;
		break;

	case BAS_S_FIELD:
		os << Indent() << "// Field Statement" << endl <<
			Indent() << "{" << endl;
		Level++;
		os << Indent() << "char* FieldBase = BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].BufferLoc();" << endl;
		Tree[1]->OutputField(os);
		Level--;
		os << Indent() << "}" << endl;
		break;

	case BAS_S_FIND:
		if (Tree[1] != NNULL)
		{
			Tree[1]->OutputGetPutOptions(Tree[0], os);
		}
		os << Indent() << "BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Find()" << endl;
		break;

	case BAS_S_FOR:
		os << Indent() << "for (";

		//
		// from
		//
		os << Tree[0]->Expression() << "; ";

		//
		// to
		//
		//	We have to try to look at the step and guess if
		//	it is a negitive step value. If we can't tell,
		//	assume positive.
		//
		if ((Tree[2] != NNULL) &&
			(Tree[2]->Type == BAS_N_UMINUS))
		{
			os << Tree[0]->Tree[0]->Expression() << " >= " <<
				Tree[1]->Expression() << "; ";
		}
		else
		{
			os << Tree[0]->Tree[0]->Expression() << " <= " <<
				Tree[1]->Expression() << "; ";
		}

		//
		// step
		//
		if (Tree[2] != NNULL)
		{
			//
			// Loop going backwards?
			//
			if (Tree[2]->Type == BAS_N_UMINUS)
			{
				os << Tree[0]->Tree[0]->Expression() << " -= " <<
					Tree[2]->Tree[0]->Expression();
			}
			else
			{
				os << Tree[0]->Tree[0]->Expression() << " += " <<
					Tree[2]->Expression();
			}
		}
		else
		{
			os << Tree[0]->Tree[0]->Expression() << "++";
		}

		os << ")" << endl;

		Block[1]->OutputBlock(os);

		break;

	case BAS_N_FORUNTIL:

		//
		// from
		//
		os << Indent() << "for (" << Tree[0]->NoParen() << "; ";

		//
		// until
		//
		os << "!(" << Tree[0]->NoParen() << ")" << "; ";

		//
		// step
		//
		os << Tree[0]->Tree[0]->Expression() << "++)" << endl;

		Block[1]->OutputBlock(os);

		break;

	case BAS_N_FORWHILE:

		//
		// from
		//
		os << Indent() << "for (" << Tree[0]->Expression() << "; ";

		//
		// until
		//
		os << Tree[0]->Tree[0]->Expression() << "==" <<
			Tree[1]->Expression() << "; ";

		//
		// step
		//
		os << Tree[0]->Tree[0]->Expression() << "++)" << endl;

		Block[1]->OutputBlock(os);

		break;

	case BAS_S_FUNCTION:
	case BAS_S_DEF:
	case BAS_S_DEFSTAR:

		//
		// Empty variable list
		//
		assert(Tree[1] != 0);

		KeepGosubFlag = GosubFlag;
		GosubFlag = 0;
		KeepOnErrorFlag = OnErrorFlag;
		OnErrorFlag = 0;

		Variables->NewLevel();

		//
		// Scan the call parameters
		//
		VariableScanOne(1);

		//
		// Scan local code
		//
		if (Block[1] != NNULL)
		{
			Block[1]->ScanForDefstar(0);
			Block[1]->VariableScan(0);
		}
		Variables->Fixup();

		//
		// Scan def* code
		//
		if (Block[2] != NNULL)
		{
			//
			// Output the DEF* code
			//
			Block[2]->OutputCode(os);
		}

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl << OutputNewDefinition() << endl;

		//
		// Braces starting function
		//
		os << Indent() << "{" << endl;
		Level++;

		//
		// Variable to hold result code
		//
		//	NOTE: Need to generate actual data type here.
		//
		if (Tree[0] != NNULL)
		{
			os << Indent() << Tree[0]->OutputNodeVarType() <<
				" Result;" << endl;
		}
		else
		{
			os << Indent() <<
				OutputVarType(GuessVarType(Tree[1]->TextValue)) <<
				" Result;" << endl;
		}

		//
		// Output local variables
		//
		if (Block[1] != 0)
		{
			Variables->OutputDef(os, Level);
			os << endl;

			if (GosubFlag != 0)
			{
				os << Indent() << "BStack(20);" << endl;
				GosubFlag = 0;
			}

			if (OnErrorFlag != 0)
			{
				os << Indent() << "OnErrorStack;" << endl;
				OnErrorFlag = 0;
			}

			Block[1]->OutputCode(os);
		}

		//
		// Give back a result code
		//
		os << Indent() << "return Result;" << endl;

		Level--;
		os << Indent() << "}" << endl;

		Variables->KillLevel();
		GosubFlag = KeepGosubFlag;
		OnErrorFlag = KeepOnErrorFlag;

		break;

	case BAS_S_MAINFUNCTION:

		//
		// Empty variable list
		//
		assert(Tree[2] != 0);

		KeepGosubFlag = GosubFlag;
		GosubFlag = 0;
		KeepOnErrorFlag = OnErrorFlag;
		OnErrorFlag = 0;

		Variables->NewLevel();

		//
		// Scan the call parameters
		//
		VariableScanOne(1);
		if (Block[1] != NNULL)
		{
			Block[1]->ScanForDefstar(0);
			Block[1]->VariableScan(0);
		}
		Variables->Fixup();

		//
		// Scan for any local def* functions
		//
		if (Block[2] != NNULL)
		{
			Block[2]->OutputCode(os);
		}

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl;
		os << Indent() << "int basic_main()" << endl;

		//
		// Braces starting function
		//
		os << Indent() << "{" << endl;
		Level++;

		//
		// Output local variables
		//
		if (Block[1] != 0)
		{
			Variables->OutputDef(os, Level);
			os << endl;

			if (GosubFlag != 0)
			{
				os << Indent() << "BStack(20);" << endl;
				GosubFlag = 0;
			}

			if (OnErrorFlag != 0)
			{
				os << Indent() << "OnErrorStack;" << endl;
				OnErrorFlag = 0;
			}

			Block[1]->OutputCode(os);
		}

		//
		// Give back a result code
		//
		os << endl << Indent() << "return EXIT_SUCCESS;" << endl;
		Level--;
		os << Indent() << "}" << endl;

		//
		// Clean up
		//
		Variables->KillLevel();
		GosubFlag = KeepGosubFlag;
		OnErrorFlag = KeepOnErrorFlag;

		break;

	case BAS_S_FUNCTIONEXIT:
		os << Indent() << "return Result;" << endl;
		break;

	case BAS_S_GET:
		if (Tree[1] != NNULL)
		{
			Tree[1]->OutputGetPutOptions(Tree[0], os);
		}
		os << Indent() << "BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Get();" << endl;

		break;

	case BAS_S_GOSUB:
		os << Indent() << "BGosub(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_GOTO:
		os << Indent() << "goto " << Tree[0]->Expression() << ";" << endl;
		break;

	case BAS_S_CONTINUE:
		if (Tree[0] != NNULL)
		{
			os << Indent() << "goto " << Tree[0]->Expression() << ";" << endl;
		}
		else
		{
			os << Indent() << "continue; // clear error, exit handler" << endl;
		}
		break;

	case BAS_S_IF:
		assert(Tree[0] != NNULL);

		os << Indent() << "if (" << Tree[0]->NoParen() << ")" << endl;

		Block[1]->OutputBlock(os);

		if (Block[2] != NNULL)
		{
			os << Indent() << "else" << endl;

			Block[2]->OutputBlock(os);
		}
		break;

	case BAS_N_WHENERRORIN:
		os << Indent() << "try" << endl;

		Block[1]->OutputBlock(os);

		if (Block[2] != NNULL)
		{
			os << Indent() << "catch(BasicError Be)" << endl;

			Block[2]->OutputBlock(os);
		}
		break;

	case BAS_P_INCLUDE:
		//
		// Don't output '#include' if we are also outputting the
		// code that was included.
		//
		if (CompileFlag == 0)
		{
			os << "#include ";
		}
		else
		{
			os << "// #include ";
		}
		os << Tree[0]->Expression();

		//
		// How should this really be handled?
		//
		if (Tree[1] != NNULL)
		{
			os << " %library " << Tree[1]->Expression();
		}
		os << endl;

		break;

	case BAS_S_INPUT:
		OutputInput(os, 1);
		break;

	case BAS_S_ITERATE:
		os << Indent();
		if (Tree[0] == NNULL)
		{
			os << "continue;" << endl;
		}
		else
		{
			os << "goto " << Tree[0]->Expression() <<
				";\t// ??? Iterate" << endl;
		}
		break;

	case BAS_S_KILL:
		os << Indent() << "unlink(" <<
			Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_V_LABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar != VNULL)
		{
			if ((ThisVar->GetEverUsed() != 0) || (KeepAllLines == TRUE))
			{
				os << ThisVar->GetName() << ":;" << endl;
			}
		}
		else
		{
			os << TextValue << ":;" << endl;
		}
		break;

	case BAS_S_LET:
		os << Indent() << Tree[0]->Expression() << ";" << endl;
		break;

	case BAS_S_LINPUT:
		OutputInput(os, 2);
		break;

	case BAS_S_LSET:
		assert(Tree[0] != NNULL);
		assert(Tree[0]->Type == BAS_N_ASSIGN);
		assert(Tree[0]->Tree[0] != NNULL);
		assert(Tree[0]->Tree[1] != NNULL);

		os << Indent() << Tree[0]->Tree[0]->Expression() << ".Lset(" <<
			Tree[0]->Tree[1]->Expression() << ");" << endl;
		break;

	case BAS_S_MAP:
		OutputMap(os);
		break;

	case BAS_S_MAT:
		assert(Tree[0] != NNULL);
		assert(Tree[1] != NNULL);

		switch(Tree[1]->Type)
		{
		case BAS_S_CON:     // MAT xxx = CON
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatCon(" << ThisVar->GetName() << ");" << endl;
			break;

		case BAS_S_IDN:
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatIdn(" << ThisVar->GetName() << ");" << endl;
			break;

		case BAS_S_ZER:
			assert(Tree[0]->Tree[0] != NNULL);
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			assert(ThisVar != 0);
			os << Indent() << "MatZer(" << ThisVar->GetName() << ");" << endl;
			break;

		case BAS_S_TRN:		// MAT xxx = TRN(yyy)
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatTrn(" << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->TextValue, Tree[1]->Tree[0]);
			os << ThisVar->GetName() << ");" << endl;
			break;

		case BAS_S_INV:		// MAT xx = INV(yyy)
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatInv(" << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->TextValue, Tree[1]->Tree[0]);
			os << ThisVar->GetName() << ");" << endl;
			break;

        case '+':			// MAT xxx = yyy + zzz
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatPlus(" << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[0]->TextValue,
				Tree[1]->Tree[0]->Tree[0]);
			os << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[1]->TextValue,
				Tree[1]->Tree[1]->Tree[0]);
			os << ThisVar->GetName() << ");" << endl;
			break;

        case '-':
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatMinus(" << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[0]->TextValue,
				Tree[1]->Tree[0]->Tree[0]);
			os << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[1]->TextValue,
				Tree[1]->Tree[1]->Tree[0]);
			os << ThisVar->GetName() << ");" << endl;
			break;

        case '*':
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatTimes(" << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[0]->TextValue,
				Tree[1]->Tree[0]->Tree[0]);
			os << ThisVar->GetName() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[1]->TextValue,
				Tree[1]->Tree[1]->Tree[0]);
			os << ThisVar->GetName() << ");" << endl;
			break;

        case '(':
			ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[0]->Tree[0]);
			os << Indent() << "MatScalar(" << ThisVar->GetName() << ", " <<
				Tree[1]->Tree[0]->Expression() << ", ";
			ThisVar = Variables->Lookup(Tree[1]->Tree[1]->TextValue,
				Tree[1]->Tree[1]->Tree[0]);
			os << ThisVar->GetName() << ");" << endl;
			break;

		}
		break;

	case BAS_N_ONERROR:
		os << Indent();
		if (atoi(Tree[0]->TextValue.c_str()) == 0)
		{
			os << "OnErrorZero;" << endl;
		}
		else
		{
			os << "OnErrorGoto(" << Tree[0]->Expression() << ");" << endl;
		}
		break;

	case BAS_N_ONGOTO:
		os << Indent() << "switch(" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			")" << endl << Indent() << "{" << endl;
		Tree[1]->OutputOngo(os, "goto", 1);
		if (Tree[2] != NNULL)
		{
				os << Indent() << "default:" << endl <<
					Indent() << "\tgoto " << Tree[2]->OutputLabel() <<
					";" << endl;
		}
		else
		{
				os << Indent() << "default:" << endl <<
					Indent() <<
					"\tthrow BasicError(58);\t// Out of range" << endl;
		}
		os << Indent() << "}" << endl;
		break;

	case BAS_N_ONGOSUB:
		os << Indent() << "switch(" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			")" << endl << Indent() << "{" << endl;
		Tree[1]->OutputOngo(os, "BGosub", 1);
		if (Tree[2] != NNULL)
		{
				os << Indent() << "default:" << endl <<
					Indent() << "\tBGosub(" <<
					Tree[2]->OutputLabel() << ");" << endl;
		}
		else
		{
			os << Indent() << "default:" << endl <<
				Indent() <<
				"\tthrow BasicError(58);\t// Out of range" << endl;
		}
		os << Indent() << "}" << endl;
		break;

	case BAS_S_OPEN:
		//
		// For INPUT/OUTPUT
		//
		if (Tree[1] != NNULL)
		{
			os << Indent() << "BasicChannel[" <<
				Tree[2]->OutputForcedType(VARTYPE_INTEGER) << "].";
			if (Tree[1]->Type == BAS_S_INPUT)
			{
				os << "ForInput();" << endl;
			}
			else
			{
				os << "ForOutput();" << endl;
			}

		}

		//
		// Other flags set on open
		//
		if (Tree[3] != NNULL)
		{
			Tree[3]->OutputOpenStuff(os, Tree[2]);
		}

		os << Indent() << "BasicChannel[" <<
			Tree[2]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Open(";
		if (Tree[0] != NNULL)
		{
			os << Tree[0]->Expression();
		}
		os << ");" << endl;
		break;

	case BAS_S_PRINT:
		OutputPrint(os);
		break;

	case BAS_S_PUT:
		if (Tree[1] != NNULL)
		{
			Tree[1]->OutputGetPutOptions(Tree[0], os);
		}
		os << Indent() << "BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Put();" << endl;
		break;

	case BAS_S_RANDOM:
		os << Indent() << "srand(time(0));" << endl;
		break;

	case BAS_S_RECORD:
		os << Indent() << "struct ";
		if (Tree[0] != NNULL)
		{
			os << Tree[0]->Expression();
		}
		os << endl << Indent() << "{" << endl;

		if (Block[1] != NNULL)
		{
			Level++;
			for (Node *loop = Block[1]; loop != NNULL; loop = loop->Block[0])
			{
				loop->OutputDefinitionList(os, NNULL, 0);
			}
			Level--;
		}

		os << Indent() << "};" << endl;

		break;

	case BAS_S_RESTORE:
	case BAS_S_RESET:
		if (Tree[0] == NNULL)
		{
			os << Indent() << "DataList.Reset();" << endl;
		}
		else
		{
			if (Tree[1] != NNULL)
			{
				Tree[1]->OutputGetPutOptions(Tree[0], os);
			}
			os << Indent() << "BasicChannel[" <<
				Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
				"].Reset();" << endl;
		}
		break;

	case BAS_S_READ:
		os << Indent() << "DataList.Read(" << Tree[0]->Expression() <<
			");" << endl;
		break;

	case BAS_S_RSET:
		assert(Tree[0] != NNULL);
		assert(Tree[0]->Type == BAS_N_ASSIGN);
		assert(Tree[0]->Tree[0] != NNULL);
		assert(Tree[0]->Tree[1] != NNULL);

		os << Indent() << Tree[0]->Tree[0]->Expression() << ".Rset(" <<
			Tree[0]->Tree[1]->Expression() << ");" << endl;
		break;

	case BAS_S_RESUME:
		os << Indent();
		if (Tree[0] != NNULL)
		{
			os << "OnErrorZero;" << endl <<
				Indent() << "goto " << Tree[0]->Expression() << ";" << endl;
		}
		else
		{
			os << "OnResume;" << endl;
		}
		break;

	case BAS_S_RETRY:
// FIXME: How to do a RETRY in C++
		os << Indent() << "retry;" << endl;
		break;

	case BAS_S_RETURN:
		os << Indent() << "BReturn;" << endl << endl;
		break;

	case BAS_S_SELECT:
		assert(Tree[0] != NNULL);

		os << Indent() << "switch (" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) << ")" << endl;

		Block[1]->OutputBlock(os);

		break;

	case BAS_S_STOP:
		os << Indent() << "exit(0);" << endl;
		break;

	case BAS_S_SUB:

		//
		// Empty variable list
		//
		assert(Tree[1] != 0);

		KeepGosubFlag = GosubFlag;
		GosubFlag = 0;
		KeepOnErrorFlag = OnErrorFlag;
		OnErrorFlag = 0;

		Variables->NewLevel();

		//
		// Scan the call parameters
		//
		VariableScanOne(1);

		//
		// Scan local code
		//
		if (Block[1] != NNULL)
		{
			Block[1]->ScanForDefstar(0);
			Block[1]->VariableScan(0);
		}
		Variables->Fixup();

		//
		// Scan def* code
		//
		if (Block[2] != NNULL)
		{
			//
			// Output the DEF* code
			//
			Block[2]->OutputCode(os);
		}

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl << OutputNewDefinition() << endl;

		//
		// Braces starting function
		//
		os << Indent() << "{" << endl;
		Level++;

		//
		// Output local variables
		//
		if (Block[1] != 0)
		{
			Variables->OutputDef(os, Level);
			os << endl;

			if (GosubFlag != 0)
			{
				os << Indent() << "BStack(20);" << endl;
				GosubFlag = 0;
			}

			if (OnErrorFlag != 0)
			{
				os << Indent() << "OnErrorStack;" << endl;
				OnErrorFlag = 0;
			}

			Block[1]->OutputCode(os);
		}

		Level--;
		os << Indent() << "}" << endl;

		Variables->KillLevel();
		GosubFlag = KeepGosubFlag;
		OnErrorFlag = KeepOnErrorFlag;

		break;

	case BAS_S_HANDLER:

		//
		// Empty variable list
		//
		assert(Tree[1] != 0);

		KeepGosubFlag = GosubFlag;
		GosubFlag = 0;
		KeepOnErrorFlag = OnErrorFlag;
		OnErrorFlag = 0;

		Variables->NewLevel();

		//
		// Scan the call parameters
		//
		VariableScanOne(1);

		//
		// Scan local code
		//
		if (Block[1] != NNULL)
		{
			Block[1]->ScanForDefstar(0);
			Block[1]->VariableScan(0);
		}
		Variables->Fixup();

		//
		// Scan def* code
		//
		if (Block[2] != NNULL)
		{
			//
			// Output the DEF* code
			//
			Block[2]->OutputCode(os);
		}

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl << OutputNewDefinition() << endl;

		//
		// Braces starting function
		//
		os << Indent() << "{" << endl;
		Level++;

		//
		// Output local variables
		//
		if (Block[1] != 0)
		{
			Variables->OutputDef(os, Level);
			os << endl;

			if (GosubFlag != 0)
			{
				os << Indent() << "BStack(20);" << endl;
				GosubFlag = 0;
			}

			if (OnErrorFlag != 0)
			{
				os << Indent() << "OnErrorStack;" << endl;
				OnErrorFlag = 0;
			}

			Block[1]->OutputCode(os);
		}

		Level--;
		os << Indent() << "}" << endl;

		Variables->KillLevel();
		GosubFlag = KeepGosubFlag;
		OnErrorFlag = KeepOnErrorFlag;

		break;

	case BAS_S_SUBEXIT:
		os << Indent() <<  "return;" << endl;
		break;

	case BAS_P_TITLE:
	case BAS_P_SBTTL:
	case BAS_P_IDENT:
	case BAS_S_REMARK:
		os << Indent();
		OutputRemark(os);
		break;

	case BAS_S_TO:
		//
		// TO used in a case statement
		//
		os << Tree[0]->Expression() << " ... " << Tree[1]->Expression();
		break;

	case BAS_S_UNLESS:
		assert(Tree[0] != NNULL);

		os << Indent();
		LookDown = Tree[0];
		while(LookDown->Type == '(')
		{
			assert(Tree[0] != NNULL);
			LookDown = LookDown->Tree[0];
		}

		//
		// Try to generate the unless as the opposite 'if' if
		// possible, otherwise just create an ungly not ('!').
		//
		switch(LookDown->Type)
		{
		case '=':
			os << "if (" << LookDown->Tree[0]->Expression() << " != " <<
				LookDown->Tree[1]->Expression() << ")" << endl;
			break;

		case '>':
			os << "if (" << LookDown->Tree[0]->Expression() << " <= " <<
				LookDown->Tree[1]->Expression() << ")" << endl;
			break;

		case '<':
			os << "if (" << LookDown->Tree[0]->Expression() << " >= " <<
				LookDown->Tree[1]->Expression() << ")" << endl;
			break;

		case BAS_X_NEQ:
			os << "if (" << LookDown->Tree[0]->Expression() << " == " <<
				LookDown->Tree[1]->Expression() << ")" << endl;
			break;

		case BAS_X_GE:
			os << "if (" << LookDown->Tree[0]->Expression() << " < " <<
				LookDown->Tree[1]->Expression() << ")" << endl;
			break;

		case BAS_X_LE:
			os << "if (" << LookDown->Tree[0]->Expression() << " > " <<
				LookDown->Tree[1]->Expression() << ")" << endl;
			break;

		case BAS_X_LNOT:
			os << "if (" << LookDown->Tree[0]->Expression() << ")" << endl;
			break;

		default:
			os << "if (!(" << LookDown->Expression() << "))" << endl;
			break;
		}

		Block[1]->OutputBlock(os);

		break;

	case BAS_S_UNLOCK:
		os << Indent() << "BasicChannel[" <<
			Tree[0]->OutputForcedType(VARTYPE_INTEGER) <<
			"].Unlock();" << endl;
		break;

	case BAS_S_UNTIL:
		assert(Tree[0] != NNULL);
		assert(Block[1] != NNULL);

		os << Indent() << "while (!(" << Tree[0]->NoParen() <<
			"))" << endl;

		Block[1]->OutputBlock(os);

		break;

	case BAS_S_UPDATE:
		os << Indent() << "RmsUpdate(" << Tree[0]->NoParen() <<
			");" << endl;
		break;

	case BAS_S_WAIT:
	case BAS_S_SLEEP:
		os << Indent() << "sleep(" << Tree[0]->NoParen() <<
			");" << endl;
		break;

	case BAS_S_WHILE:
		assert(Tree[0] != NNULL);
		assert(Block[1] != NNULL);

		os << Indent() << "while (" << Tree[0]->NoParen() <<
			")" << endl;

		Block[1]->OutputBlock(os);
		break;

	default:
		os << "??%" << Type << "@" << lineno << ":" << TextValue << "(";
		if (Tree[0] != NNULL)
		{
			os << Tree[0];
		}
		os << ",";
		if (Tree[1] != NNULL)
		{
			os << Tree[1];
		}
		os << ",";
		if (Tree[2] != NNULL)
		{
			os << Tree[2];
		}
		os << ",";
		if (Tree[3] != NNULL)
		{
			os << Tree[3];
		}
		os << ")" << endl;
		break;
	}
}


/*******************************************************************************

Local Function:

	Node::Expression()

Title:

	Generates string value of an exporeesion.

Description:

	This function will convert an expression into a string, ready
	for output.

Bugs:

	This used to be part of OutputCodeOne, but has been broken out to
	make understanding of the code easier, and to return a string
	value instead of outputing to a channel.

*******************************************************************************/

string Node::Expression()
{
	string result;
	VariableStruct *ThisVar;

	assert(this != NNULL);

	//
	// Handle this node type
	//
	switch (Type)
	{
	case BAS_N_ASSIGNLIST:
		assert(Tree[0] != NNULL);
		assert(Tree[1] != NNULL);

		if (Tree[0]->Type == BAS_V_FUNCTION)
		{
			result = "Result = ";
		}
		else
		{
			result = Tree[0]->Expression() + " = ";
		}

		if (Tree[1]->Type == BAS_V_FUNCTION)
		{
			result += "Result";
		}
		else
		{
			result += Tree[1]->Expression();
		}
		break;

	case BAS_V_USELABEL:
		result = OutputLabel();
		break;

	case BAS_V_FUNCTION:
		//
		// Look for variable in table
		//
		ThisVar = Variables->Lookup(TextValue, Tree[0]);

		//
		// Print this variable
		//
		if (ThisVar != VNULL)
		{
			if (ThisVar->GetName() == "Tab")
			{
				if (IOChannel == NNULL)
				{
					result = "BasicChannel[0].";
				}
				else
				{
					result = string("BasicChannel[") +
						IOChannel->Expression() + "].";
				}
			}
			result += ThisVar->GetName();
		}
		else
		{
			result = TextValue;
		}

		if (Tree[0] != NNULL)
		{
			result += "(" + Tree[0]->NoParen() + ")";
		}
		else
		{
			result += "()";
		}
		break;

	case BAS_N_FORASSIGN:
		result = Tree[0]->Expression() + " = " + Tree[1]->Expression();
		break;

	case BAS_X_STRREF:
		result = Tree[0]->Expression() + "." + Tree[1]->Expression();
		break;

	case BAS_N_STRUCTNAME:
		result = TextValue;
		break;

	case BAS_N_NULL:
		result = "NULL";
		break;

	//
	// I/O Channel
	//
	case '#':
		result = Tree[0]->OutputForcedType(VARTYPE_INTEGER);
		break;

	//
	// Equality statement
	//
	case '=':
		result = Tree[0]->OutputBstringText() + " == " +
			Tree[1]->Expression();
		break;

	case '+':
	case '-':
	case '*':
	case '/':
	case '>':
	case '<':
		result = Tree[0]->OutputBstringText() + " " + TextValue +
			" " + Tree[1]->Expression();
		break;

	case BAS_S_MOD:
		result = Tree[0]->Expression() + " % " + Tree[1]->Expression();
		break;

	case '(':
		//
		// Lose unecessary additional levels of pren's
		//
		result = string("(") + Tree[0]->NoParen() + ")";
		break;

	case '^':
		result = string("pow(") + Tree[0]->NoParen() + ", " +
			Tree[1]->NoParen() + ")";
		break;

	case BAS_X_LAND:
		result = Tree[0]->Expression() + " && " + Tree[1]->Expression();
		break;

	case BAS_S_AND:
		result = Tree[0]->Expression() + " & " + Tree[1]->Expression();
		break;

	case BAS_S_BY:
		switch(Tree[1]->Type)
		{
		case BAS_S_DESC:
			result = string("$descr(") + Tree[0]->Expression() + ")";
			break;

		case BAS_S_REF:
			result = string("&") + Tree[0]->Expression();
			break;

		default:
			result = Tree[0]->Expression();
			break;
		}
		break;

	case BAS_S_EQV:
		result = string("BasicEqv( ") + Tree[0]->Expression() + ", " +
			Tree[1]->Expression() + ")";
		break;

	case BAS_S_IMP:
		result = string("BasicImp( ") + Tree[0]->Expression() + ", " +
			Tree[1]->Expression() + ")";
		break;

	case BAS_X_GE:
		result = Tree[0]->Expression() + " >= " + Tree[1]->Expression();
		break;

	case BAS_V_INT:
	case BAS_V_INTEGER:
	case BAS_V_FLOAT:
		result = OutputInteger();
		break;

	case BAS_X_LE:
		result = Tree[0]->Expression() + " <= " + Tree[1]->Expression();
		break;

	case BAS_N_LIST:
		if (Tree[0] == NNULL)
		{
			result = string("0") + ", ";
		}
		else
		{
			result = Tree[0]->Expression() + ", ";
		}

		if (Tree[1] == NNULL)
		{
			result += "0";
		}
		else
		{
			result += Tree[1]->Expression();
		}
		break;

	case BAS_V_NAME:
		result = OutputVarName(Tree[0]);
		break;

	case BAS_X_EQ:
		result = Tree[0]->Expression() + " == " + Tree[1]->Expression();
		break;

	case BAS_X_NEQ:
		result = Tree[0]->Expression() + " != " + Tree[1]->Expression();
		break;

	case BAS_S_NOT:
		result = string("~") + Tree[0]->Expression();
		break;

	case BAS_X_LNOT:
		result = string("!") + Tree[0]->Expression();
		break;

	case BAS_X_LOR:
		result = Tree[0]->Expression() + " || " + Tree[1]->Expression();
		break;

	case BAS_S_OR:
		result = Tree[0]->Expression() + " | " + Tree[1]->Expression();
		break;

	case BAS_V_PREDEF:
		result = TextValue;
		break;

	case BAS_V_TEXTSTRING:
		result = TextValue;
		break;

	case BAS_N_TYPEFUN:
		result = string("(") + Tree[1]->OutputNodeVarType() +
			")(" + Tree[0]->Expression() + ")";
		break;

	case BAS_N_UMINUS:
		result =  string("-") + Tree[0]->Expression();
		break;

	case BAS_N_UPLUS:
		result = string("+") + Tree[0]->Expression();
		break;

	case BAS_X_UNARYGT:
		result = string("> ") + Tree[0]->Expression();
		break;

	case BAS_X_UNARYGE:
		result = string(">= ") + Tree[0]->Expression();
		break;

	case BAS_X_UNARYLE:
		result = string("<= ") + Tree[0]->Expression();
		break;

	case BAS_X_UNARYLT:
		result = string("< ") + Tree[0]->Expression();
		break;

	case BAS_X_UNARYNEQ:
		result = string("!= ") + Tree[0]->Expression();
		break;

	case BAS_S_XOR:
		result = Tree[0]->Expression() + " ^ " + Tree[1]->Expression();
		break;

	default:
		{
			// FIXME: There must be a better way
			char buffer[32];
			sprintf(buffer, "%d@%d", Type, lineno);
			result = string("[Unknown Node Type ") + buffer + "]";
		}
		break;
	}

	return result;
}

/*******************************************************************************

Member Function:

	Node::OutputVarName()

Title:

	Outputs Variable Name

Description:

	Creates the full name for a variable

*******************************************************************************/

string Node::OutputVarName(Node* array, int InDefine)
{
	//
	// Local Variables
	//
	VariableStruct *ThisVar;
	string result;

	assert(this != NNULL);

	//
	// Look for variable in table
	//
	ThisVar = Variables->Lookup(TextValue, array);

	//
	// Make up a good name to print for this variable
	//
	if (ThisVar != VNULL)
	{
		result = ThisVar->GetName();
	}
	else
	{
		result = TextValue;
	}

	//
	// Handle array references (May need to be modified from x[a,b]
	// to x[a][b] manually.
	//
	if (array != NNULL)
	{
		if ((ThisVar != VNULL) && (ThisVar->GetClass() == VARCLASS_FUNC))
		{
			result += "(" + array->NoParen() + ")";
		}
		else
		{
			while(array != NNULL)
			{
				result += "[" + array->OutputArrayParam(InDefine) + "]";
				array = array->Block[0];
			}
		}
	}

	return result;
}

/*******************************************************************************

Member Function:

	Node::OutputPrint()

Title:

	Outputs a print statement.

Description:

	This function will mangle a print statement.

Bugs:


*******************************************************************************/

void Node::OutputPrint(ostream& os)
{
	int ReturnFlag = 0;

	assert(this != NNULL);

	//
	// Indent as necessary
	//
	IOChPrinted = 0;	// IO Channel not yet output
	IOChannel = NNULL;	// No IO Channel defined yet
	IOUsing = NNULL;		// No Print Using Statement yet

	//
	// Output print using
	//
	if (Tree[0] != NNULL)
	{
		//
		// Handle print statement
		//
		ReturnFlag = Tree[0]->OutputPrintData(os, 0);

		//
		// Finish off print using statement
		//
		if (IOUsing != NNULL)
		{
			OutputIPChannel(os, 0);
			os << " << PUse.Finish()";
		}

		//
		// Force a return at end of line?
		//
		if (ReturnFlag != 0)
		{
			if ((IOChannel == NNULL) || (IOChPrinted != 0))
			{
				OutputIPChannel(os, 0);
				os << " << endl";
			}
			else
			{
				//
				// When using a BChannel, it doesn't handle this
				// case properly in C++, so fake it.
				//
				OutputIPChannel(os, 0);
				os << ".GetOstream() << endl";
			}
		}
	}
	else
	{
		//
		// Empty print command
		//
		if ((IOChannel == NNULL) || (IOChPrinted != 0))
		{
			OutputIPChannel(os, 0);
			os << " << endl";
		}
		else
		{
			//
			// When using a BChannel, it doesn't handle this
			// case properly in C++, so fake it.
			//
			OutputIPChannel(os, 0);
			os << ".GetOstream() << endl";
		}
	}

	//
	// Close off statement
	//
	if (IOChPrinted != 0)
	{
		os << ";" << endl;
	}
}

//******************************************************************************
//
void Node::OutputIPChannel(ostream& os, int InputFlag)
{
	//
	// Output channel only if we haven't already
	//
	if (IOChPrinted == 0)
	{
		os << Indent();

		//
		// Use default if no channel specified, else use
		// whatever is given
		//
		if (IOChannel == NNULL)
		{
			os << "BasicChannel[0]";
		}
		else
		{
			os << "BasicChannel[" << IOChannel->Expression() << "]";
		}

		//
		// Flag that we have printed the channel
		//
		IOChPrinted = 1;
	}
}

/*******************************************************************************

Member Function:

	Node::OutputInput()

Title:

	Outputs a input statement.

Description:

	This function will mangle a input statement.

Bugs:


*******************************************************************************/

void Node::OutputInput(ostream& os, int InputFlag)
{
	assert(this != NNULL);

	//
	// Initialize flags
	//
	IOChPrinted = 0;		// Channel not yet printed
	IOChannel = NNULL;		// No channel specified
	IOUsing = NNULL;			// No print using specified

	//
	// Write out input statements
	//
	if (Tree[0] != NNULL)
	{
		Tree[0]->OutputInputData(os, InputFlag);
	}

	//
	// Finish off input statement
	//
	if (IOChPrinted != 0)
	{
		os << ";" << endl;
	}
}


/*******************************************************************************

Member Function:

	int Node::OutputPrintData()

Title:

	Outputs data related to an input or print statement.

Description:

	This function will mangle a input statement. Returns a
	flag depending on if the input ended in something that should
	cause a return to be appended.

Parameters:

	InputFlag

		Specifies if it is working on an input statement,
		ie. specifies the direction of the arrows.

Return:

	0 - Blank item on right hand side.
	1 - Normal item on right hand side.
	2 - Special item was just output

Bugs:


*******************************************************************************/

int Node::OutputPrintData(ostream& os, int InputFlag)
{
	int ReturnFlag;		// Does a return seem to be needed
				//   Gets set to one if the print/Input command
				//   ends with a ',' or ';'.
	int TestFlag = 0;	// Test for special item

	assert(this != NNULL);

	//
	// Is this a piece of a list?
	//
	switch (Type)
	{
	case BAS_N_LIST:
		//
		// Try the left side of the tree
		//
		if (Tree[0] != NNULL)
		{
			TestFlag = Tree[0]->OutputPrintData(os, InputFlag);
		}

		//
		// Do we need to append a tab between fields?
		// Only do so for print statements.
		//
		if ((TextValue[0] == ',') &&
			(TestFlag != 2) && (IOUsing == NNULL))
		{
			OutputIPChannel(os, InputFlag);
			os << " << '\\t'";
		}

		//
		// Handle right side of the tree
		//
		if (Tree[1] != NNULL)
		{
			ReturnFlag = Tree[1]->OutputPrintData(os, InputFlag);
		}
		else
		{
			ReturnFlag = 0;
		}

		break;

	case '#':
		IOChannel = this;
		ReturnFlag = 2;
		break;

	case BAS_S_USING:
		IOUsing = Tree[0];
		ReturnFlag = 2;

		os << Indent() << "PUse.SetFormat(" << IOUsing->Expression() <<
			");" << endl;

		break;

	case BAS_N_RECORD:
		OutputIPChannel(os, InputFlag);
		os << " << RECORD(" << Tree[0]->Expression() << ")";
		ReturnFlag = 2;
		break;

	case BAS_V_TEXTSTRING:
		OutputIPChannel(os, InputFlag);

		//
		// Put a comma in print using statements, else
		// out out a '<<'.
		//
		if (IOUsing != NNULL)
		{
			os << " << PUse.Output(" << Expression() << ");" << endl;
			IOChPrinted = 0;
		}
		else
		{
			os << " << " << Expression();
		}
		ReturnFlag = 1;

		break;


	default:

		//
		// Print
		//
		if (CheckCpos())
		{
			//
			// If we have an item that needs the current
			// cursor position to be correct, we have to put
			// it at tyhe beginning of an output statement
			//
			os << ";" << endl;
			IOChPrinted = 0;
		}
		OutputIPChannel(os, InputFlag);

		//
		// Put a comma in print using statements, else
		// out out a '<<'.
		//
		if (IOUsing != NNULL)
		{
			os << " << PUse.Output(" << Expression() << ");" << endl;
			IOChPrinted = 0;
		}
		else
		{
			os << " << " << Expression();
		}
		ReturnFlag = 1;
		break;
	}

	//
	// Done
	//
	return ReturnFlag;
}

/*******************************************************************************

Member Function:

	int Node::CheckCpos()

Title:

	Check for functions that depend on cursor position

Description:

	This function is used to check for functions that are sensitive
	to cursor position, since all of the parameters in a 'cout <<'
	are evaluated before any output is done.

*******************************************************************************/

int Node::CheckCpos()
{
	//
	// Check this node
	//
	if (Type == BAS_V_FUNCTION)
	{
		VariableStruct *ThisVar;
		ThisVar = Variables->Lookup(TextValue, Tree[0]);

		//
		// Print this variable
		//
		if (ThisVar != VNULL)
		{
			string name = ThisVar->GetName();
			if ((name == "Tab") || (name == "Ccpos"))
			{
				return 1;
			}
		}
	}

	//
	// Check any parameters
	//
	for (int loop = 0; loop <= 4; loop++)
	{
		if ((Tree[loop] != NNULL) && (Tree[loop]->CheckCpos()))
		{
			return 1;
		}
	}

	//
	// Most not be anything
	//
	return 0;
}

/*******************************************************************************

Member Function:

	int Node::OutputInputData()

Title:

	Outputs data related to an input or print statement.

Description:

	This function will mangle a input statement. Returns a
	flag depending on if the input ended in something that should
	cause a return to be appended.

Parameters:

	InputFlag

		Specifies if it is working on an input statement,
		ie. specifies the direction of the arrows.

Return:

	0 - Blank item on right hand side.
	1 - Normal item on right hand side.
	2 - Special item was just output

Bugs:


*******************************************************************************/

int Node::OutputInputData(ostream& os, int InputFlag)
{
	int ReturnFlag;		// Does a return seem to be needed
				//   Gets set to one if the print/Input command
				//   ends with a ',' or ';'.
	int TestFlag = 0;	// Test for special item

	assert(this != NNULL);

	//
	// Is this a piece of a list?
	//
	switch (Type)
	{
	case BAS_N_LIST:
		//
		// Try the left side of the tree
		//
		if (Tree[0] != NNULL)
		{
			TestFlag = Tree[0]->OutputInputData(os, InputFlag);
		}

		//
		// Handle right side of the tree
		//
		if (Tree[1] != NNULL)
		{
			ReturnFlag = Tree[1]->OutputInputData(os, InputFlag);
		}
		else
		{
			ReturnFlag = 0;
		}

		break;

	case '#':
		IOChannel = this;
		ReturnFlag = 2;
		break;

	case BAS_S_USING:
		IOUsing = Tree[0];
		ReturnFlag = 2;

		os << Indent() << "PUse.SetFormat(" << IOUsing->Expression() <<
			");" << endl;

		break;

	case BAS_N_RECORD:
		OutputIPChannel(os, InputFlag);
		os << " >> RECORD(" << Tree[0]->Expression() << ")";
		ReturnFlag = 2;
		break;

	case BAS_V_TEXTSTRING:
		//
		// If this is a prompt used for input, close the current
		// input and set up an output prompt.
		//
		if (IOChPrinted != 0)
		{
			os << ";" << endl;
		}

		//
		// If this is an input statement, create the whole
		// statement on one line
		//
		os << Indent() << "BasicChannel[0] << " <<
			Expression() << ";" << endl;
		ReturnFlag = 2;
		IOChPrinted = 0;
		break;


	default:

		//
		// Output seperator depending on what state we are in
		//
		switch(InputFlag)
		{
		//
		// Regular input
		//
		case 1:
			OutputIPChannel(os, InputFlag);
			os << " >> " << Expression();
			ReturnFlag = 1;
			break;

		//
		// LINPUT/INPUT LINE
		//
		case 2:
			os << Indent() << "{" << endl;
			Level++;
			os << Indent() << "char TmpInput[MAX_INPUT];" << endl;
			OutputIPChannel(os, InputFlag);
			os << ".getline(TmpInput, MAX_INPUT);" << endl <<
				Indent() << Expression() << " = TmpInput;" << endl;
			Level--;
			os << Indent() << "}" << endl;
			IOChPrinted = 0;
			ReturnFlag = 2;
			break;
		}

		break;
	}

	//
	// Done
	//
	return ReturnFlag;
}

/*******************************************************************************

Global Function:

	Node::OutputData()

Title:

	Outputs a data statement.

Description:

	This function will mangle a data statement.

Bugs:


*******************************************************************************/

static int DataWidth;

void Node::OutputData(ostream& os)
{
	assert(this != NNULL);

	DataWidth = 0;
	os << Indent();

	//
	// Handle NULL values specially
	//
	if (Tree[0] == NNULL)
	{
		os << "\"\"";
	}
	else
	{
		Tree[0]->OutputDataValue(os);
	}

	//
	// Handle additional data statements attached to this one
	//
	Node* Loop = Tree[1];
	while (Loop != NNULL)
	{
		//
		// Let's start on a new line for each new
		// data statement, so we keep more of the
		// original formatting.
		//
		os << "," << endl;
		DataWidth = 0;
		os << Indent();

		//
		// Handle NULL values specially
		//
		if (Loop->Tree[0] == NNULL)
		{
			os << "\"\"";
		}
		else
		{
			Loop->Tree[0]->OutputDataValue(os);
		}
		Loop = Loop->Tree[1];
	}
}


/*******************************************************************************

Local Function:

	Node::OutputDataValue()

Title:

	Outputs one data value in a data statement.

Description:

	This function will mangle one data value for a data statement.
	It forces everything to be a string.

Bugs:


*******************************************************************************/

void Node::OutputDataValue(ostream& os)
{
	assert(this != NNULL);

	if (Type == BAS_N_LIST)
	{
		//
		// Handle NULL values specially
		//
		if (Tree[0] == NNULL)
		{
			os << "\"\"";
		}
		else
		{
			Tree[0]->OutputDataValue(os);
		}
		os << ",";
		DataWidth++;
		if (DataWidth > 60)
		{
			os << endl;
			DataWidth = 0;
			os << Indent();
		}
		//
		// Handle NULL values specially
		//
		if (Tree[1] == NNULL)
		{
			os << "\"\"";
		}
		else
		{
			Tree[1]->OutputDataValue(os);
		}
	}
	else
	{
		if (Type == BAS_V_TEXTSTRING)
		{
			os << Expression();
			DataWidth += TextValue.length();
		}
		else
		{
			//
			// Force quotes around data value
			//
			os << '"' << Expression()  << '"';
			DataWidth += TextValue.length() + 2;
		}
	}
}

/*******************************************************************************

Global Function:

	Node::OutputMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/

void Node::OutputMap(ostream& os)
{
	assert(this != NNULL);
	assert(Tree[0] != NNULL);

	VariableStruct* ThisVar = Variables->Lookup(Tree[0]->TextValue, NNULL);
	assert(ThisVar != VNULL);

	os << Indent() << "struct " << ThisVar->GetName() << " ";

	if (Tree[1] != NNULL)
	{
		os << endl << Indent() << "{" << endl;

		Level++;
		Tree[1]->OutputDefinitionList(os, NNULL, 0);
		Level--;

		os << Indent() << "} ";
	}
	os << ThisVar->GetName() << ";" << endl;
}

/*******************************************************************************

Member Function:

	Node::OutputCaseLabel()

Title:

	Output the "case" selections, one at a time.

Description:


Bugs:


*******************************************************************************/

void Node::OutputCaseLabel(ostream& os)
{
	assert(this != NNULL);

	//
	// Handle this if it is a list
	//
	if (Type == BAS_N_LIST)
	{
		Tree[0]->OutputCaseLabel(os);
		Tree[1]->OutputCaseLabel(os);
		return;
	}

	//
	// Output one case statement
	//
	Level--;
	os << Indent() << "case " << Expression() << ":" << endl;
	Level++;
}


/*******************************************************************************

Local Function:

	OutputRemark()

Title:

	Prints out a comment.

Description:

	This function strips off all of the junk from a comment
	and prints it out.

Bugs:


*******************************************************************************/

void Node::OutputRemark(ostream& os)
{
	string Remark = TextValue;

	//
	// Skip start of remark character
	//
	if (Remark[0] == '!')
	{
#ifdef STLREMOVE
		Remark.remove(0,1);
#else
		Remark.erase(0,1);
#endif
	}

	//
	// Skip extra white space
	//
	if (Remark[0] == ' ')
	{
#ifdef STLREMOVE
		Remark.remove(0,1);
#else
		Remark.erase(0,1);
#endif
	}

	//
	// Lose any junk at end of comment
	// (Yes, it does modify the original value)
	//
	while(Remark.length() > 0)
	{
		// Line continuation marker
		if (*(Remark.end() - 1) == '&')
		{
#ifdef STLREMOVE
			Remark.remove(Remark.end() - 1);
#else
			Remark.erase(Remark.end() - 1);
#endif
			continue;
		}

		// New Line
		if (*(Remark.end() - 1) == '\n')
		{
#ifdef STLREMOVE
			Remark.remove(Remark.end() - 1);
#else
			Remark.erase(Remark.end() - 1);
#endif
			continue;
		}

		// Space
		if (*(Remark.end() - 1) == ' ')
		{
#ifdef STLREMOVE
			Remark.remove(Remark.end() - 1);
#else
			Remark.erase(Remark.end() - 1);
#endif
			continue;
		}

		// Tab
		if (*(Remark.end() - 1) == '\t')
		{
#ifdef STLREMOVE
			Remark.remove(Remark.end() - 1);
#else
			Remark.erase(Remark.end() - 1);
#endif
			continue;
		}

		// Exclimation point
		if (*(Remark.end() - 1) == '!')
		{
#ifdef STLREMOVE
			Remark.remove(Remark.end() - 1);
#else
			Remark.erase(Remark.end() - 1);
#endif
			continue;
		}

		// Break out of loop
		break;
	}

	//
	// Dump out remark (and maybe add a space to the front)
	//
	if ((Remark.length() == 0) || (Remark[0] == '\t') || (Remark[0] == '*'))
	{
		os << "//" << Remark;
	}
	else
	{
		os << "// " << Remark;
	}

	//
	// There may be some extra stuff that needs printing
	//
	if (Tree[0] != NNULL)
	{
		os << ' ' << Tree[0]->Expression();
	}

	os << endl;

}

/*******************************************************************************

Function:

	OutputOngo()

Title:

	Prints out a computed goto/gosub statement

Description:

	This translates an 'ON xxx GOTO yyy,zzz,...' into a select
	statement. The 'switch' portion is generated before calling this
	code, this only outputs the case labels/goto's.

	It returns nonzero if a default case was output, else it returns 0.

Bugs:


*******************************************************************************/

int Node::OutputOngo(ostream& os, const char* OnType, int GoLevel)
{
	int flag = 0;

	switch(Type)
	{
	case BAS_N_LIST:
		flag = Tree[0]->OutputOngo(os, OnType, GoLevel);
		flag += Tree[1]->OutputOngo(os, OnType, GoLevel + 1);
		break;

	case BAS_V_USELABEL:
		os << Indent() << "case " << GoLevel << ":" << endl;
		Level++;
		os << Indent() << OnType;
		if (strcmp(OnType, "goto") == 0)
		{
			os << ' ' << OutputLabel() << ";" << endl;
			// goto doesn't need to be followed by a break
		}
		else
		{
			os << '(' << OutputLabel() << ");" << endl;
			os << Indent() << "break;" << endl;
		}
		Level--;
		break;

	default:
		os << "Error in computed " << OnType <<
			" (" << Expression() << ") @ " << lineno << endl;
		break;
	}
	return flag;
}

/*******************************************************************************

Function:

	OutputLabel()

Title:

	Prints out a label

Description:

	Output a label (if referenced anywhere)

Bugs:


*******************************************************************************/

string Node::OutputLabel()
{
	VariableStruct *ThisVar = Variables->Lookup(TextValue, 0);
	if (ThisVar != VNULL)
	{
		return ThisVar->GetName();
	}
	else
	{
		return TextValue;
	}
}

/*******************************************************************************

Function:

	OutputDefinitionList()

Title:

	Prints out a series of definitions

Description:


Bugs:


*******************************************************************************/

void Node::OutputDefinitionList(ostream& os, Node* MainType, int ExtType, int GlobalStat)
{
	assert(this != NNULL);

	switch(Type)
	{
	case BAS_N_LIST:

cerr << "LIST still used in definition" << endl;
		assert(Tree[0] != NNULL);
		assert(Tree[1] != NNULL);

		//
		// Put multiple items on seperate lines
		//
		Tree[0]->OutputDefinitionList(os, MainType, ExtType);
		Tree[1]->OutputDefinitionList(os, MainType, ExtType);

		break;

	case BAS_V_DEFINEVAR:
		os << Indent();

		switch(GlobalStat)
		{
		case 1:		// Constant definition
			os << "const ";
			break;

		case 3:		// External constant
			os << "extern const ";
			break;

		case 4:		// External
			os << "extern ";
			break;
		}

		if (Tree[1] != NNULL)
		{
			os << Tree[1]->OutputNodeVarType() << " ";
		}
		if (Tree[3] != NNULL)
		{
			os << Tree[3]->OutputPassmech(0);
		}
		if (Tree[0] != NNULL)
		{
			os << Tree[0]->OutputVarName(Tree[2], 1);
		}
		if (Tree[4] != NNULL)
		{
			os << " = " << Tree[4]->Expression();
		}

		os << ";" << endl;

		if (Block[0] != NNULL)
		{
			Block[0]->OutputDefinitionList(os, MainType, ExtType, GlobalStat);
		}
		break;

	case BAS_V_DEFINEFUN:
		assert(Tree[0] != NNULL);

		os << Indent();

		switch(GlobalStat)
		{
		case 2:		// Constant definition
			os << "extern ";
			break;
		}

		if (Tree[1] != NNULL)
		{
			os << Tree[1]->OutputNodeVarType() << " ";
		}
		if (Tree[3] != NNULL)
		{
			os << Tree[3]->OutputPassmech(1);
		}
		os << Tree[0]->OutputVarName(Tree[2]);
		if (Tree[4] != NNULL)
		{
			os << " = ";
			Tree[4]->OutputCodeOne(os);
		}
		os << ";" << endl;
		if (Block[0] != NNULL)
		{
			Block[0]->OutputDefinitionList(os, MainType, ExtType, GlobalStat);
		}
		break;

	case BAS_S_VARIANT:

		os << Indent() << "union" << endl <<
			Indent() << "{" << endl;
		Level++;
		Block[1]->OutputDefinitionList(os, MainType, ExtType, GlobalStat);
		Level--;
		os << Indent() << "};" << endl;

		break;

	case BAS_S_CASE:
		//
		// Variant case
		//
		for (Node* loop = this; loop != NNULL; loop = loop->Block[0])
		{
			os << Indent() << "struct" << endl <<
				Indent() << "{" << endl;
			Level++;
			loop->Block[1]->OutputDefinitionList(os, MainType, ExtType,
				GlobalStat);
			Level--;
			os << Indent() << "};" << endl;
		}

		break;


	default:
		os << Indent() << OutputDefinition(MainType, ExtType) << ";" << endl;
		break;
	}
}

/*******************************************************************************

Function:

	OutputVirtualList()

Title:

	Prints out a series of Virtual Array definitions

Description:


Bugs:


*******************************************************************************/
void Node::OutputVirtualList(ostream& os, Node* Channel,
	Node* MainType, int ExtType, int GlobalStat)
{
	VariableStruct *ThisVar;

	assert(this != NNULL);
	assert(Channel != NNULL);

	switch(Type)
	{
	case BAS_N_LIST:

		assert(Tree[0] != NNULL);
		assert(Tree[1] != NNULL);

		//
		// Put multiple items on seperate lines
		//
		Tree[0]->OutputVirtualList(os, Channel, MainType, ExtType);
		Tree[1]->OutputVirtualList(os, Channel, MainType, ExtType);

		break;

	default:
		//
		// Look for variable in table
		//
		assert(Tree[0] != NNULL);
		assert(Tree[2] != NNULL);
		assert(Channel != NNULL);

		ThisVar = Variables->Lookup(Tree[0]->TextValue, Tree[2]);

		os << Indent() <<
			"VirtualArray<" <<
			Tree[1]->OutputNodeVarType() << "> ";

		if (ThisVar != 0)
		{
			os << ThisVar->GetName() << "(";
		}
		else
		{
			os << Tree[0]->TextValue << "(";
		}
		os << Channel->Expression() << ", " << Tree[2]->Expression();

		if (Tree[3] != NNULL)
		{
			os << ", " << Tree[3]->Expression();
		}
		os << ");" << endl;

		if (Block[0] != NNULL)
		{
			Block[0]->OutputVirtualList(os, Channel, MainType, ExtType);
		}
		break;

	}
}

/*******************************************************************************

Function:

	OutputDefinitionParams()

Title:

	Prints out a series of paramater definitions

Description:


Bugs:


*******************************************************************************/

string Node::OutputDefinitionParams(Node* MainType, int ExtType, int GlobalStat)
{
	assert(this != NNULL);
	string result;
	assert(Type != BAS_N_LIST); // Should no longer happen

	switch(Type)
	{
	case BAS_V_DEFINEVAR:
		switch(GlobalStat)
		{
		case 1:		// Constant definition
			result = "const ";
			break;
		}

		if (Tree[1] != NNULL)
		{
			result += Tree[1]->OutputNodeVarType() + " ";
		}
		if (Tree[3] != NNULL)
		{
			result += Tree[3]->OutputPassmech(0);
		}
		if (Tree[0] != NNULL)
		{
			result += Tree[0]->OutputVarName(Tree[2], 1);
		}
		if (Tree[4] != NNULL)
		{
			result += " = " + Tree[4]->Expression();
		}

		if (Block[0] != NNULL)
		{
			result += ", " +
				Block[0]->OutputDefinitionParams(MainType, ExtType, GlobalStat);
		}
		break;

	case BAS_V_DEFINEFUN:
		assert(Tree[0] != NNULL);

		if (Tree[1] != NNULL)
		{
			result += Tree[1]->OutputNodeVarType() + " ";
		}
		if (Tree[3] != NNULL)
		{
			result += Tree[3]->OutputPassmech(1);
		}
		result += Tree[0]->OutputVarName(Tree[2]);
		if (Tree[4] != NNULL)
		{
			result += " = " + Tree[4]->Expression();
		}
		if (Block[0] != NNULL)
		{
			result += ", " +
					Block[0]->OutputDefinitionParams(MainType, ExtType, GlobalStat);
		}
		break;

	default:
 cerr << "Something isn't using BAS_V_DEFINE??? (default) @ " << lineno << endl;
		result += "const " + OutputDefinition(MainType, ExtType);
		break;
	}

	return result;
}


/*******************************************************************************

 OutputPassMech

	Handle writing out the parameters to a function/definition/var

*******************************************************************************/

string Node::OutputPassmech(int FunctionFlag)
{
	string result;

	switch(Type)
	{
	case BAS_S_VALUE:
		break;

	case BAS_S_REF:
		result = "&";
		break;

	case BAS_S_DESC:
		result = "& /* Descriptor */ ";
		break;
	}

	return result;
}

/*******************************************************************************

Function:

	OutputNewDefinition()

Title:

	Prints out a definition

Description:

	This function will output a function/subroutine definition
	in a format that is acceptable to C++.

Bugs:

	Doesn't yet know about passing mechanisms

*******************************************************************************/

string Node::OutputNewDefinition()
{
	VariableStruct *ThisVar;
	int IsFunction = 0;
	string result;

	assert(this != NNULL);
	assert(Tree[1] != NNULL);

	//
	// Handle class and type
	//
	switch(Type)
	{
	case BAS_N_EXTERNALFUNCTION:
		assert(Tree[0] != NNULL);
		IsFunction = 1;
		result = string("extern ") + Tree[0]->OutputNodeVarType() + " ";
		break;

	case BAS_N_EXTERNALSUB:
		assert(Tree[0] == NNULL);
		IsFunction = 1;
		result = "extern void ";
		break;

	case BAS_S_FUNCTION:
		assert(Tree[0] != NNULL);
		IsFunction = 1;
		result = Tree[0]->OutputNodeVarType() + " ";
		break;

	case BAS_S_DEF:
	case BAS_S_DEFSTAR:
		IsFunction = 1;
		result = "static ";
		if (Tree[0] != NNULL)
		{
			result += Tree[0]->OutputNodeVarType() + " ";
		}
		else
		{
			result += OutputVarType(GuessVarType(Tree[1]->TextValue)) + " ";
		}
		break;

	case BAS_N_EXTERNALCONSTANT:
		assert(Tree[0] != NNULL);
		result = "extern const " + Tree[0]->OutputNodeVarType() + " ";
		break;

	case BAS_S_SUB:
		IsFunction = 1;
		result = "void ";
		break;

	case BAS_S_HANDLER:
		IsFunction = 1;
		result = "void ";
		break;
	}

	//
	// Look for variable in table
	//
	ThisVar = Variables->Lookup(Tree[1]->TextValue, Tree[3]);

	//
	// Make up a good name to print for this variable
	//
	if (ThisVar != VNULL)
	{
		result += ThisVar->GetName();
	}
	else
	{
		result += "?" + Tree[1]->TextValue + "?";
	}

	if (Tree[3] != NNULL)
	{
		if (IsFunction == 0)
		{
			//
			// Handle references (Function or array)
			// (May need to be modified from x[a,b]
			// to x[a][b] manually.
			//
			if (Tree[3] != NNULL)
			{
				if (Tree[3]->IsSimpleInteger())
				{
					// FIXME: Isn't there a better way?
					char buffer[16];
					sprintf(buffer, "[%d]",
						(atoi(Tree[3]->TextValue.c_str()) + 1));
					result += buffer;
				}
				else
				{
					result += "[" +
						Tree[3]->OutputForcedType(VARTYPE_INTEGER) + " + 1]";
				}
			}
		}
		else
		{
			//
			// Handle references (Function)
			//
			result += "(";

			if (Tree[3] != NNULL)
			{
				result += Tree[3]->OutputDefinitionParams(NNULL, BAS_N_LIST);
			}
			else
			{
				result += "...";
			}

			result += ")";
		}
	}
	else
	{
		if (IsFunction != 0)
		{
			result += "(void)";
		}
	}

	//
	// Output any constant definitions
	//
	if (Tree[4] != NNULL)
	{
		result += " = " + Tree[4]->Expression();
	}

	return result;
}

/*******************************************************************************

Function:

	OutputDefinition()

Title:

	Prints out a definition

Description:


Bugs:


*******************************************************************************/

string Node::OutputDefinition(Node* MainType, int ExtType)
{
	VariableStruct *ThisVar;
	string result;

	assert(this != NNULL);

	switch(Type)
	{
	case BAS_N_ASSIGN:

		result = Tree[0]->OutputDefinition(MainType, BAS_N_ASSIGN) +
			" = " + Tree[1]->Expression();
		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:
	case BAS_N_STRUCTNAME:

		//
		// This is a data type, so see if there are any variables
		// listed under this type, else dump out the data type
		// all be itself.
		//
		if (Tree[2] != NNULL)
		{
			result = Tree[2]->OutputDefinition(this, ExtType);
		}
		else
		{
			result = OutputNodeVarType() + " ";
		}
		if (Tree[3] != NNULL)
		{
			if (Tree[3]->IsSimpleInteger())
			{
				// FIXME: There has to be a better way
				char buffer[16];
				sprintf(buffer, "[%d]", (atoi(Tree[3]->TextValue.c_str()) + 1));
				result += buffer;
			}
			else
			{
				result += "[" +
					Tree[3]->OutputForcedType(VARTYPE_INTEGER) + " + 1]";
			}
		}
		break;

	default:
		//
		// Handle class
		//
		switch(ExtType)
		{
		case BAS_S_EXTERNAL:
			result = "extern ";
			break;

		case BAS_N_EXTERNALCONSTANT:
			result = "extern const ";
			break;

		}

		switch (Type)
		{
		case BAS_V_NAME:
			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != VNULL)
			{
				result += OutputVarType(ThisVar->GetType()) + " ";
			}
			else
			{
				if (MainType != NNULL)
				{
					result += OutputVarType(MainType->Type) + " ";
				}
				else
				{
					result += "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			if (ThisVar != VNULL)
			{
				result += ThisVar->GetName();
			}
			else
			{
				result += TextValue;
			}

			//
			// Handle references (Function or array)
			// (May need to be modified from x[a,b]
			// to x[a][b] manually.
			//
			if (Tree[0] != NNULL)
			{
				result += "[" + Tree[0]->OutputArrayParam(1) + "]";
			}
			if (Tree[3] != NNULL)
			{
				if (Tree[3]->IsSimpleInteger())
				{
					// FIXME: There has to be a better way
					char buffer[16];
					sprintf(buffer, "[%d]",
						(atoi(Tree[3]->TextValue.c_str()) + 1));
					result += buffer;
				}
				else
				{
					result += "[" +
						Tree[3]->OutputForcedType(VARTYPE_INTEGER) + " + 1]";
				}
			}
			break;

		case BAS_V_FUNCTION:

			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != VNULL)
			{
				result += OutputVarType(ThisVar->GetType()) + " ";
			}
			else
			{
				if (MainType != NNULL)
				{
					result += OutputVarType(MainType->Type) + " ";
				}
				else
				{
					result += "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			if (ThisVar != VNULL)
			{
				result += ThisVar->GetName();
			}
			else
			{
				result += TextValue;
			}

			//
			// Handle references (Function)
			//
			result += "(";

			if (Tree[0] != NNULL)
			{
				result += Tree[0]->OutputDefinitionParams(NNULL, BAS_N_LIST);
			}
			else
			{
				if ((ExtType == BAS_S_EXTERNAL) ||
					(ExtType == BAS_N_EXTERNALSUB) ||
					(ExtType == BAS_N_EXTERNALFUNCTION))
				{
					result += "...";
				}
				else
				{
					result += "void";
				}
			}

			result += ")";

			break;

		default:
			{
				// FIXME: There has to be a better way
				char buffer[32];
				sprintf(buffer, "%d@%d", Type, lineno);
				result = string("(??) ") + buffer + "(??)" + TextValue;
			}
			break;
		}
		break;

	}

	return result;
}

/*******************************************************************************

Function:

	OutputArrayParam()

Title:

	Prints out an array reference list

Description:


Bugs:


*******************************************************************************/

string Node::OutputArrayParam(int Increment)
{
	assert(this != NNULL);

	string result;

	switch(Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NNULL)
		{
			result = Tree[0]->OutputArrayParam(Increment);
		}
		else
		{
			result = "(NULL)";
		}
		result += "][";
		if (Tree[1] != NNULL)
		{
			result += Tree[1]->OutputArrayParam(Increment);
		}
		else
		{
			result += "(NULL)";
		}
		break;

	default:

		if (Increment == 1)
		{
			if (IsSimpleInteger())
			{
				// FIXME: There really ought to be a better way
				char buffer[16];
				sprintf(buffer, "%d", atoi(TextValue.c_str()) + 1);
				result = buffer;
			}
			else
			{
				result = OutputForcedType(VARTYPE_INTEGER) + " + 1";
			}
		}
		else
		{
			result = OutputForcedType(VARTYPE_INTEGER);
		}
		break;

	}
	return result;
}


/*******************************************************************************

Member Function:

	OutputOpenStuff()

Title:

	Output code for an 'OPEN' statement

Description:

	This function generates the code necessary to handle
	a full 'RMS' open statement. It generates calls to a class
	for all the elements available in the statement.

*******************************************************************************/

void Node::OutputOpenStuff(ostream& os, Node* Channel)
{
	switch(Type)
	{
	case BAS_S_RECORDSIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetRecordSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_CLUSTERSIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetClusterSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_FILESIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER);
		os << "].SetFileSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_EXTENDSIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetExtendSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_WINDOWSIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetWindowSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_BLOCKSIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetBlockSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_BUCKETSIZE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetBucketSize(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_MODE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetMode(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_BUFFER:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetBuffer(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_DEFAULTNAME:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetDefaultName(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_PRIMARY:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetKey(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_ALTERNATE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetKey(" << Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_TEMPORARY:

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetTemporary(True);" << endl;
		break;

	case BAS_S_CONTIGUOUS:

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetContiguous(True);" << endl;
		break;

	case BAS_S_NOREWIND:

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetRewind(False);" << endl;
		break;

	case BAS_S_NOSPAN:

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetSpan(False);" << endl;
		break;

	case BAS_S_SPAN:

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetSpan(True);" << endl;
		break;

	case BAS_S_ORGANIZATION:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetOrginization(" << Tree[0]->TextValue;
		if (Tree[0]->Tree[0] != NNULL)
		{
			os << ", " << Tree[0]->Tree[0]->TextValue;
		}
		os << ");" << endl;
		break;

	case BAS_S_ACCESS:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetAccess(" << Tree[0]->TextValue << ");" << endl;
		break;

	case BAS_S_ALLOW:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetAllow(" << Tree[0]->TextValue << ");" << endl;
		break;

	case BAS_S_CONNECT:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetConnect(BasicChannel[" << Tree[0]->Expression() << "]);" <<
			endl;
		break;

	case BAS_S_MAP:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetMap(" << Tree[0]->Expression() << ", sizeof(" <<
			Tree[0]->Expression() <<
			"));" << endl;
		break;

	case BAS_S_RECORDTYPE:

		assert(Tree[0] != NNULL);

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetRecordType(" << Tree[0]->TextValue <<
			");" << endl;
		break;

	case BAS_S_VIRTUAL:
		break;

	default:

		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].GoofyOpenStuff(" << Expression() << ");" << endl;
		break;

	}

	//
	// Handle downlink
	//
	if (Block[0] != NNULL)
	{
		Block[0]->OutputOpenStuff(os, Channel);
	}
}

/*******************************************************************************

Member Function:

	OutputClose()

Title:

	Output code for a 'CLOSE' statement

Description:

	This function generates the code necessary to handle
	a full 'RMS' close statement.

*******************************************************************************/

void Node::OutputClose(ostream& os)
{
	switch(Type)
	{
	case BAS_N_LIST:
		Tree[0]->OutputClose(os);
		Tree[1]->OutputClose(os);
		break;

	default:
		os << Indent() << "BasicChannel[" <<
			OutputForcedType(VARTYPE_INTEGER) <<
			"].Close();" << endl;
		break;
	}
}

/***********************************************************************

OutputBstringText

	Force first parameter to be a 'BString' instead of a string constant.

Description:

	This function is used to force the first item in a string expression
	to be a BString, so that the compiler won't choke on exporessions
	like "ABC"+"DEF".
************************************************************************/
string Node::OutputBstringText()
{
	string result;

	switch(Type)
	{
	case BAS_V_TEXTSTRING:
		result = "BString(" + Expression() + ")";
		break;

	default:
		result = Expression();
		break;
	}

	return result;
}

/***********************************************************************
 * Member Function:
 *
 *	Node::OutputInteger
 *
 * Title:
 *
 *	Output an integer/number
 *
 * Description:
 *
 *	Used to output integers, while stripping off unecessary stuff
 *	that confuses C++, like leading zeros and trailing percent
 *	signs..
 */
string Node::OutputInteger()
{
	int OFlag = 0;		// Have we output something yet?
	int Ptr = 0;		// What character are we looking at
	char ch;		// Actual character
	string result;		// Working string

	//
	// Output all digits in the string
	//
	while ((ch = TextValue[Ptr]), (Ptr < TextValue.length()))
	{
		Ptr++;
		if (ch == '0')
		{
			if (OFlag != 0)
			{
				result += ch;
			}
		}
		else
		{
			if (ch != '%')
			{
				if ((ch == '.') && (OFlag == 0))
				{
					result += '0';
				}
				result += ch;
				OFlag = 1;
			}
		}
	}

	//
	// If we didn't output anything, then output a zero
	// because it must have been all zeros.
	//
	if (OFlag == 0)
	{
		result += '0';
	}

	return result;
}


/***********************************************************************
 * Member Function:
 *
 *	Node::OutputField
 *
 * Title:
 *
 *	Output a field statement
 *
 * Description:
 *
 *	Used to generate code for field statements
 */
void Node::OutputField(ostream& os)
{
	assert(this != NNULL);

	switch(Type)
	{
	case BAS_N_LIST:

		assert(Tree[0] != NNULL);
		assert(Tree[1] != NNULL);

		Tree[0]->OutputField(os);
		Tree[1]->OutputField(os);

		break;

	case BAS_S_AS:
		os << Indent() << Tree[1]->Expression() <<
			".ReMap(FieldBase, " << Tree[0]->Expression() <<
			");" << endl <<
			Indent() << "FieldBase += " << Tree[0]->Expression() << ";" << endl;
		break;

	default:
		os << "// **** This shouldn't be possible ***** @ " << lineno << endl <<
			Indent() << TextValue;
		break;
	}
}

/***********************************************************************
 * OutputGetPutOptions
 *
 *	Output code got get/put options
 *
 * Description:
 *
 *	Outputs code for get/put/find for such options as 'record',
 *	'rfa', 'regardless' and other modifiers.
 */
void Node::OutputGetPutOptions(Node* Channel, ostream& os)
{
	switch(Type)
	{
	case BAS_N_LIST:
		if (Tree[0] != NNULL)
		{
			Tree[0]->OutputGetPutOptions(Channel, os);
		}
		if (Tree[1] != NNULL)
		{
			Tree[1]->OutputGetPutOptions(Channel, os);
		}
		break;

	case BAS_V_RMSRECORD:
	case BAS_S_RECORD:
	case BAS_S_BLOCK:
		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetRecord(" <<
			Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_REGARDLESS:
		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetRegardless();" << endl;
		break;

	case BAS_S_KEY:
		os << Indent() << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetKey(" << Tree[0]->Expression() << ");" << endl;
		if (Tree[1] != NNULL)
		{
			os << Indent() << "BasicChannel[" <<
				Channel->OutputForcedType(VARTYPE_INTEGER) <<
				"].SetKeyMode(";
			switch(Tree[1]->Type)
			{
			case BAS_S_EQ:
				os << "Equal";
				break;
			case BAS_S_GE:
				os << "GreaterEqual";
				break;
			case BAS_S_GT:
				os << "Greater";
				break;

			}
			os << ");" << endl <<
				Indent() << "BasicChannel[" <<
				Channel->OutputForcedType(VARTYPE_INTEGER) <<
				"].SetKeyValue(" << Tree[2]->Expression() << ");" << endl;
		}
		break;

	case BAS_S_RFA:
		os << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetRfaValue(" <<
			Tree[0]->Expression() << ");" << endl;
		break;

	case BAS_S_COUNT:
		os << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].SetCount(" << Tree[0]->Expression() << ");" << endl;
		break;
	default:
		os << "BasicChannel[" <<
			Channel->OutputForcedType(VARTYPE_INTEGER) <<
			"].Goofy(" << Expression() << ");" << endl;
	}
}

/***********************************************************************
 * OutputForcedType
 *
 *	Calculate the varible type for a given node.
 *
 * Description:
 *
 *	Determines the variable type for a given node, and returns
 *	the specified type.
 */
string Node::OutputForcedType(int VarType)
{
	int ThisType = GetNodeVarType();	// Var type of this node
	int NeedForce = 0;			// Do we need to force the type
	string result;

	//
	// A NULL will match anything
	//
	if (Type == BAS_N_NULL)
	{
		result = NoParen();
		return result;
//		return 0;
	}

	//
	// Determine if we really need to force
	//
	switch(VarType)
	{
	case VARTYPE_INTEGER:
		//
		// Integer type allows for BYTE, INTEGER, LONG
		//
		if ((ThisType != VARTYPE_BYTE) &&
			(ThisType != VARTYPE_WORD) &&
			(ThisType != VARTYPE_INTEGER) &&
			(ThisType != VARTYPE_LONG))
		{
			NeedForce = 1;
		}
		break;

	default:
		if (ThisType != VarType)
		{
			NeedForce = 1;
		}
	}

	//
	// Now force out the proper code
	//
	if (NeedForce)
	{
		result = string("(") + OutputVarType(VarType) + ")(" +
			NoParen() + ")";
	}
	else
	{
		result = NoParen();
	}

	return result;
}

/***********************************************************************
 * OutputNodeVarType
 *
 *	Output the variable type definition for a node
 *
 * Description:
 *
 *	This function is used to output the variable type of a type node
 *	in a variable declaration statement.
 */
string Node::OutputNodeVarType()
{
	string result;

// cerr << "Called OutputNodeVarType Type=" << Type  << ", Text=" << TextValue << endl;
	switch(Type)
	{
	case BAS_N_STRUCTNAME:
	case BAS_N_VARTYPE:
// FIXME: This needs attention, at least lower case the bloody name.
		result = TextValue;
		break;

	default:
		result = OutputVarType(Type);
		break;
	}

	return result;
}

