/*******************************************************************************

Program:

	variable.cc

Title:

	Routines to handle variables.

Description:

	This file contains routines to handle the variable definitions
	necessary for basic.

Bugs:

History:

	08/30/1996 - Kevin Handy
		Update version number to 1.1, and update on internet.

	10/22/1996 - Kevin Handy
		Handle new 'IntegerType' and 'RealType' variables.

	04/30/1997 - Kevin Handy
		Guestimate the number of parameters to a array/function,
		and use this guess when outputing default definitions.

	06/16/1997 - Kevin Handy
		Set version to 1.6

	06/24/1997 - Kevin Handy
		Output correct for 'VARTYPE_REAL' based on 'option' value

	07/06/1997 - Kevin Handy
		Broke out some common code into the functions DumpVarType,
		OutputVarType, and GuessVarType.

	11/28/1997 - Kevin Handy
		Removed Left, Right fields (Replaced by STL functions).
		Added copy constructor for STL.

	12/15/1997 - Kevin Handy
		Changed 'Node' member 'TextValue' from 'char*' to 'string'

	01/26/1998 - Kevin Handy
		Force basic names to upper case

	02/05/1998 - Kevin Handy
		Move several class functions into the header, greatly
		reducing executable size.

	08/20/1999 - Kevin Handy
		Added UnguessVartype function

	06/16/2000 - Kevin Handy
		Added code to GuessVartype to handle --integer option.

*******************************************************************************/

/*
 * System Include Files
 */
#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

//
// Project Include Files
//
#include "basic.h"
#include "variable.h"
#include "parse.h"


/*******************************************************************************

Member Function:

	Dump()

Title:

	Print out variable.

Description:

	Prints out a variables information.

Bugs:

	Ugly output.

*******************************************************************************/

void VariableStruct::Dump() const
{
	//
	// Print out this variable
	//
	cout << "Var: " << BasicName;

	if (Class == VARCLASS_ARRAY)
	{
		cout << "[" << ParCount << "]";
	}

	cout << ", Type " << Type << DumpVarType(Type) <<
		", Class " << Class;

	switch (Class)
	{
	case VARCLASS_NONE:
		cout << " Regular variable";
		break;

	case VARCLASS_ARRAY:
		cout << " An Array";
		break;

	case VARCLASS_FUNC:
		cout << " Function";
		break;

	case VARCLASS_CONST:
		cout << " Constant";
		break;

	case VARCLASS_MAP:
		cout << " Map";
		break;

	default:
		cout << " Undefined";
		break;
	}

	if (CName.length() != 0)
	{
		cout << ", CName " << CName;

		if (Class == VARCLASS_ARRAY)
		{
			cout << "[]";
		}
	}

	cout << ", Usage " << EverUsed;
	cout << endl;
}


/*******************************************************************************

Member Function:

	VariableStruct::OutputDef()

Title:

	Dump out one variable node.

Description:

	This function will output one variables definition to the
	output file. These are only used if the definition isn't output
	as part of the source code.

Bugs:


*******************************************************************************/

string VariableStruct::OutputDef()
{
	//
	// Local Variables
	//
	string result;
	int ActualType = Type;

	//
	// Output This Variable
	//
	switch (Class)
	{
	case VARCLASS_CONST:
		result = "const ";
		break;
	}

	result += OutputVarType(ActualType);
	result += " ";

	//
	// Pick a good name
	//
	if (CName.length() != 0)
	{
		result += CName;
	}
	else
	{
		result += BasicName;
	}

	switch(Type)
	{
	case VARTYPE_FIXSTR:
		result += "[]";
		break;
	}


	//
	// Add brackets if this is an array definition
	//
	switch (Class)
	{
	case VARCLASS_ARRAY:
		//
		// Default array dimension under Basic is 0..10
		// for most Basic's I've seen.
		//
		result += "[11]";
		for (int loop = 2; loop <= ParCount; loop++)
		{
			result += "[11]";
		}
		break;

	case VARCLASS_FUNC:
		result += "()";
		break;
	}

	//
	// Flag this one as having been output.
	//
	Output = TRUE;
	return result;
}

/*******************************************************************************

Member Function:

        GuessVarType()

Title:

        Guess the type of a variable based on it's name

Description:

        Looks at a variables name and attempts to guess what type
	of variable it is. This type of guess is only used when
	an actual definition is not given.

Bugs:

        Really simple guesses.

*******************************************************************************/

int GuessVarType(const string Name)
{
	switch(*(Name.end() - 1))
	{
	case '$':
		return VARTYPE_DYNSTR;
	case '%':
		return VARTYPE_INTEGER;
	default:
		return DefaultType;
	}
}

int UnguessVarType(int vartype)
{
	switch(vartype)
	{
	case VARTYPE_RFA:
		return BAS_S_RFA;
	case VARTYPE_FIXSTR:
	case VARTYPE_DYNSTR:
		return BAS_S_STRING;
	case VARTYPE_BYTE:
		return BAS_S_BYTE;
	case VARTYPE_WORD:
		return BAS_S_WORD;
	case VARTYPE_INTEGER:
		return BAS_S_INTEGER;
	case VARTYPE_LONG:
		return BAS_S_LONG;
	case VARTYPE_SINGLE:
		return BAS_S_SINGLE;
	case VARTYPE_REAL:
		return BAS_S_REAL;
	case VARTYPE_DOUBLE:
		return BAS_S_DOUBLE;
	case VARTYPE_HFLOAT:
		return BAS_S_HFLOAT;
	case VARTYPE_GFLOAT:
		return BAS_S_GFLOAT;
	case VARTYPE_DECIMAL:
		return BAS_S_DECIMAL;
	default:
		return BAS_S_INTEGER;
	}
}

/*******************************************************************************

Member Function:

	DumpType()

Title:

	Print out variable type.

Description:

	Prints out a variable type.

Bugs:

	Ugly output.

*******************************************************************************/

const char* DumpVarType(int VarType)
{
	char* result;

	switch (VarType)
	{
	case VARTYPE_DYNSTR:
		result = " Dynamic String";
		break;

	case VARTYPE_FIXSTR:
		result = " Fixed length string";
		break;

	case VARTYPE_WORD:
		result = " 16 bit integer";
		break;

	case VARTYPE_INTEGER:
		result = " integer";
		break;

	case VARTYPE_LONG:
		result = " 32 bit integer";
		break;

	case VARTYPE_REAL:
		result = " float";
		break;

	case VARTYPE_DOUBLE:
		result = " double";
		break;

	case VARTYPE_STRUCT:
		result = " Structure";
		break;

	case VARTYPE_VOID:
		result = " Void";
		break;

	case VARTYPE_BYTE:
		result = " Byte";
		break;

	case VARTYPE_RFA:
		result = " RFA";
		break;

	case VARTYPE_LABEL:
		result = " Label";
		break;

	case VARTYPE_DECIMAL:
		result = " Decimal";
		break;

	case VARTYPE_SINGLE:
		result = " Single";
		break;

	case VARTYPE_GFLOAT:
		result = " GFloat";
		break;

	case VARTYPE_HFLOAT:
		result = " HFloat";
		break;

	default:
		result = " Undefined Type";
		break;
	}
	return result;
}

/*******************************************************************************

Member Function:

	OutputVarType()

Title:

	Print out variable type.

Description:

	Prints out a variable type.

Bugs:

	Ugly output.

*******************************************************************************/

string OutputVarType(int VarType)
{
	int ActualType = VarType;
	char* result;

	switch(ActualType)
	{
	case VARTYPE_DYNSTR:
	case BAS_S_STRING:
		result = "BString";
		break;

	case VARTYPE_FIXSTR:
		result = "char";
		break;

	case VARTYPE_INTEGER:
	case BAS_S_INTEGER:
		switch(IntegerType)
		{
		case VARTYPE_WORD:
			result = "Int";
			break;

		case VARTYPE_LONG:
			result = "Int32";
			break;

		case VARTYPE_BYTE:
			result = "char";
			break;

		default:
			result = "??unknown int??";
			break;
		}
		break;

	case VARTYPE_WORD:
	case BAS_S_WORD:
		result = "Int";
		break;

	case VARTYPE_LONG:
	case BAS_S_LONG:
		result = "Int32";
		break;

	case VARTYPE_REAL:
	case BAS_S_REAL:
		switch(RealType)
		{
		case VARTYPE_DOUBLE:
			result = "double";
			break;

		case VARTYPE_GFLOAT:
			result = "Gfloat";
			break;

		case VARTYPE_HFLOAT:
			result = "Hfloat";
			break;

		default:
			result = "??unknown float??";
			break;
		}
		break;

	case VARTYPE_DOUBLE:
	case BAS_S_DOUBLE:
		result = "double";
		break;

	case VARTYPE_VOID:
		result = "void";
		break;

	case VARTYPE_BYTE:
	case BAS_S_BYTE:
		result = "char";
		break;

	case VARTYPE_RFA:
	case BAS_S_RFA:
		result = "RFA";
		break;

	case VARTYPE_DECIMAL:
	case BAS_S_DECIMAL:
		result = "Decimal";
		break;

	case VARTYPE_SINGLE:
	case BAS_S_SINGLE:
		result = "Sfloat";
		break;

	case VARTYPE_GFLOAT:
	case BAS_S_GFLOAT:
		result = "Gfloat";
		break;

	case VARTYPE_HFLOAT:
	case BAS_S_HFLOAT:
		result = "Hfloat";
		break;

	case VARTYPE_LABEL:
		result = "label";
		break;

	case BAS_N_STRUCTNAME:
		cerr << "FIXME: Called OutputVarType with a structure name" << endl;
		result = "struct";
		break;

	default:
		result = "??unknown??";
		break;

	}
	return string(result);
}

